/* $Workfile: GodarPCIeDemo.c $       $Revision: 9 $      $Modtime: 1/27/11 3:44p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc OtxDriverTest

  @module

    GodarPCIeDemo.c

    Application for Godar-Pci-Express board.

    Timeslot TS0 is used to store framing data for the E1 and T1. 
    The T1 interface uses all 84 highways. The E1 interface uses 63 
    highways but the highways are not contunous. That is HWY 0-21, 
    HWY 28-48 and HWY 56-76 are used to store the data to transfer 
    one Li per highway. The demo does not enable the unused highways.
    

    T1 channels are stored as follows in TS 0-31 of each highway:
      TS00   Frame bit
      TS01   TS1
      TS02   TS2
      TS03   TS3
      TS04   Not used fill data 
      TS05   TS4
      TS06   TS5
      TS07   TS6
      TS08   Not used fill data 
      TS09   TS7
      TS10   TS8
      TS11   TS9
      TS12   Not used fill data 
      TS13   TS10
      TS14   TS11
      TS15   TS12
      TS16   Not used fill data 
      TS17   TS13
      TS18   TS14
      TS19   TS15
      TS20   Not used fill data 
      TS21   TS16
      TS22   TS17
      TS23   TS18
      TS24   Not used fill data 
      TS25   TS19
      TS26   TS20
      TS27   TS21
      TS28   Not used fill data 
      TS29   TS22
      TS30   TS23
      TS31   TS24

  Note:
    Currently the FPGA supports all E1 highways and is configured
    for E1 data transfers.
    The Godar board supports the OTX_TEMUX_INTERFACE_HMVIP Temux interface.
    The Godar board supports data monitoring.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxData.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"
#include "OtxTemux.h"
#include "OtxTemuxD.h"

// Set demo for T1 configuration
//#define DEMO_USE_T1 1

#define DEMO_DEFAULT_LOG_FILE "GodarPCIeDemo.log"
#define DEMO_DEFAULT_BOARD_NO 0
#define DEMO_T1_MAX_HWY_NO    84
#ifdef DEMO_USE_T1
#define DEMO_MAX_HWY_NO       DEMO_T1_MAX_HWY_NO // 84 for each SONET/SDH span
#else
#define DEMO_MAX_HWY_NO       77 // STM-1 (E1) only uses 63 highways
#endif
#define DEMO_ACTIVE_HWY_NO    DEMO_MAX_HWY_NO
#define DEMO_GAMMA_INCREMENT  1 // Each TS will be increased by this value
// All highway generate events get all the highways
#define DEMO_EVENT_DATA_BUF_SIZE DEMO_MAX_HWY_NO + 20
#define DEMO_BURST_TS_SIZE    32
 
#ifdef DEMO_BURST_BUF_SIZE
#undef DEMO_BURST_BUF_SIZE
#endif
#define DEMO_BURST_BUF_SIZE  (DEMO_BURST_TS_SIZE *  0x200) 


typedef struct DemoEnvT {
  char                    *pszInputStr;
  OTX_BOOL                bExitOnError;
  char                    logFileName[OTX_MAX_PATH];
  OTX_INT32               nBoardType;
  OTX_UINT32              boardNo;
  OTX_UINT32              coLi;
  OTX_BOOL                bDisableBurstTx;
  OTX_BOOL                bDisableBurstRx;
  // Temux configuration
  OtxTemuxDevConfigureE   eTemuxConfig;
  OtxTemuxTxTimingE       eTxTimingMode;
  OtxTemuxMapTribModeE    eLinkSrcOrDest;
  OtxTemuxFrameModeE      eE1FramingMode;
  OtxTemuxClockSrcE       eClkSrcMode;
} DemoEnvT;

static void DemoCheckRc(IN OTX_RESULT result);
static void DemoLogf(char *formatStr, ...);
static void DemoLogfile(char *formatStr, ...); // No console display log file only
static void DemoExitQuestion(void);
static void DemoTemuxInterfaceUserDefined(void);
static void DemoBurstTestGammaData(OTX_UINT32 nHwy);
static void DemoBurstEndTest(void);
static void DemoBurstCheckRefenceId(OTX_UINT64 nFrameRefId, OTX_UINT32 nSize, OTX_UINT32 nHwy);
static void DemoBurstCheckFixedPattern(OTX_UINT32 *pInputData, OTX_UINT32 nHwy);
static void DemoCheckBurstGammaData(OTX_UINT8 *pBuf, OTX_UINT32 *pErrCnt, OTX_UINT32 nHighway);
static void DemoBurstSender(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize);
static void DemoBurstReceiver(OTX_UINT32 nSourceId, OTX_UINT32 *pBlockSize);
static void DemoConnectBurstDevices(void);
static void DemoDisconnectBurstDevices(void);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static OTX_BOOL DemoDisplayConfiguration(OTX_UINT32 nConfig);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);

static OTX_UINT8 g_sqSendFixedBuf[DEMO_T1_MAX_HWY_NO][DEMO_BURST_BUF_SIZE];
  
OtxTemuxGetTemuxInfoS g_sTemuxInformation;

OTX_UINT32 g_nPattern = 0;
OTX_UINT32 g_nSlipCnt = 0;

DemoEnvT    g_demoEnv;
FILE *g_pDemoLogFile = NULL;
OTX_EVENT   g_hNotificationEvent      = OTX_INVALID_EVENT_VALUE;
OTX_EVENT   g_hNotificationBurstEvent = OTX_INVALID_EVENT_VALUE;
// Each highway has it own event queue so no highway is blocked while processing
// the data from another highway.
OTX_EVENT   g_hBurstReceiveEvent[DEMO_T1_MAX_HWY_NO]; 
OTX_HANDLE  g_hBurstReceiveEventQueue[DEMO_T1_MAX_HWY_NO];
OTX_INT32   g_nHighwayNumber[DEMO_T1_MAX_HWY_NO];

OTX_HANDLE g_hBoard            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hTemux            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hBurst            = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hEventQueue       = OTX_INVALID_HANDLE_VALUE;
OTX_HANDLE g_hBurstEventQueue  = OTX_INVALID_HANDLE_VALUE;

// Used by the PrintStats function
OTX_UINT8  g_nFrame;
OTX_BOOL   g_nFrameFlag = OTX_FALSE;
OTX_INT32  g_nBurstExitFlag   = 0;
OTX_INT32  g_FetchCount       = 0;
OTX_INT32  g_FetchCountCur    = 0;
// Highways used by this demo.The data on the other highways is
// not processed.
OTX_INT32  g_nActiveHighways  = DEMO_ACTIVE_HWY_NO; 
OTX_UINT32 g_nLink            = 0; // Number of Temux spans supported by Godar
OTX_UINT32 g_nTstHighway      = 0; // verbose highway


OTX_BOOL g_bVerbose       = OTX_FALSE; // Enable additional print message
OTX_BOOL g_bBurstVerbose  = OTX_FALSE; // Enable data print message
OTX_BOOL g_bDemoRunnig    = OTX_TRUE;  // When false threads will exit

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;

  OTX_UINT32  nGammaZeroXferCnt;
  OTX_UINT8   sqGammaData[DEMO_BURST_TS_SIZE];
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
  OTX_INT32   nOverflowErrCnt;  
  OTX_UINT64  nFrameRefId;

  OTX_INT32   nDataBlockCounter;
  OTX_BOOL    bGammaTestMode;
  OTX_BOOL    bFixedTestMode;
  OTX_BOOL    bDisplayDataErr;
  OTX_BOOL    bTestReset;
  OTX_INT32   nGammaTestErrors[DEMO_BURST_TS_SIZE];
  OTX_UINT8   sqGammaTestData[DEMO_BURST_TS_SIZE];
};

// one for each possible highway hosted by the Burst device
struct hLogicDevs {
  struct sBurstSender mBurstSender[DEMO_T1_MAX_HWY_NO]; 
  struct sBurstReceiver mBurstReceiver[DEMO_T1_MAX_HWY_NO]; 
} g_hLogicDevs;



/*******************************************************************************
*******************************************************************************/
void DemoCheckRc(IN OTX_RESULT result)
{
  if ( OTX_FAILURE(result) ) {
    char msgStr[256];
    OTX_RESULT otxRc;
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(0, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", 
      OTX_RESULT_CODE(result), 
      OTX_RESULT_SOURCE_ID(result), 
      OTX_RESULT_LINE_NUM(result), 
      msgStr);
    
    DemoExitQuestion();
  }
}


static void DemoCheckDevRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;
  DemoLogf("Do you want to exit (Y/N) ");
  c = OsaConsoleGetCh();
  
  if (c == 'y' || c == 'Y') {
    // Clean up
    DemoLogf(" Good Bye\n");
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }

}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Only print it to the logfile, console print takes too much time

  @comm

*******************************************************************************/
static void DemoLogfile(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile, formatStr, argList);

  va_end(argList);
}


/*******************************************************************************
  @func
    Get the hardware device information from Temux
*******************************************************************************/
void DemoTemuxGetDeviceInfo(void)
{
  OTX_UINT16 nLink;
  OTX_UINT16 nMaxLink = OTX_TEMUX_MAX_LINK;
  OTX_UINT16 nSts1;
  OTX_UINT16 nSrcPort;
  OtxTemuxDestPortE eDestPort;
  OTX_UINT16 nVtGroup;
  OTX_UINT16 nTrib;
  OTX_UINT16 nParm = 0;

  OTX_RESULT otxRc;
  OTX_BOOL   bDeviceConfigured = OTX_TRUE;

  otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
  DemoCheckDevRc(g_hTemux, otxRc);

  DemoLogf("Interface %s\n", g_sTemuxInformation.eHwInterface ? "HMVIP" : "SBI");

  // For now just display this information
  DemoLogf("Number of Lines (Links) supported %d\n", g_sTemuxInformation.nNumSpans);
  if((g_sTemuxInformation.nNumSpans < OTX_TEMUX_MAX_LINK) && 
     (g_sTemuxInformation.nNumSpans != 0)) {
    nMaxLink = (OTX_UINT16) g_sTemuxInformation.nNumSpans;

    g_nLink = g_sTemuxInformation.nNumSpans;
  }
  else {
    DemoLogf("Number of Lines supported is out of range\n");
    g_nLink = OTX_TEMUX_MAX_LINK;
  }

  // OtxTemuxDevConfigureE
  DemoLogf("Configuration Mode 0x%X\n", g_sTemuxInformation.eConfigMode); 
  DemoDisplayConfiguration((OTX_UINT32) g_sTemuxInformation.eConfigMode);

  // OtxTemuxIntfModeE
  DemoLogf("SONET/SDH physical interface 0x%X - ", g_sTemuxInformation.eIntfMode); 
  switch(g_sTemuxInformation.eIntfMode) {
  case OTX_TEMUX_INTF_MODE_1_OC12_STM4:
    DemoLogf("OTX_TEMUX_INTF_MODE_1_OC12_STM4\n"); 
    break;

  case OTX_TEMUX_INTF_MODE_OC3_STM1:
    DemoLogf("OTX_TEMUX_INTF_MODE_OC3_STM1\n"); 
    break;

  default:
    DemoLogf("Not Configured OtxTemuxIntfModeE\n"); 
  }
  // OtxTemuxSts12PayloadE
  DemoLogf("SONET/SDH  STS payload 0x%X - ", g_sTemuxInformation.eSts12Payload); 
  switch(g_sTemuxInformation.eSts12Payload) {
  case OTX_TEMUX_4_STS_3:
    DemoLogf("OTX_TEMUX_4_STS_3\n"); 
    break;

  case OTX_TEMUX_STS_12:
    DemoLogf("OTX_TEMUX_STS_12\n"); 
    break;

  default:
    DemoLogf("Not ConfiguredOtxTemuxSts12PayloadE\n"); 
  }
  // OtxTemuxSysIfModeE
  DemoLogf("PDH Subsystem data path configuration 0x%X - ", 
    g_sTemuxInformation.sStsiDestSrc.eSysIfMode); 
  switch(g_sTemuxInformation.sStsiDestSrc.eSysIfMode) {
  case OTX_TEMUX_SBI:
    DemoLogf("OTX_TEMUX_SBI\n"); 
    break;

  case OTX_TEMUX_SBITR:
    DemoLogf("OTX_TEMUX_SBITR\n"); 
    break;

  case OTX_TEMUX_HMVIP:
    DemoLogf("OTX_TEMUX_HMVIP\n"); 
    break;

  default:
    DemoLogf("Not Configured OtxTemuxSysIfMode\n"); 
    bDeviceConfigured = OTX_FALSE; // It should be setup
  }
  for(nParm=0; nParm < 5; nParm++) {
    DemoLogf("Parm%d is 0x%X\n", 
             nParm+1, // API has parm range as 1-5 
             g_sTemuxInformation.sStsiDestSrc.nParam[nParm]);
  }

  DemoLogf("\n");

  for(nLink=0; nLink < nMaxLink; nLink++) {
    DemoLogf("\n");
    DemoLogf("Link %d\n", nLink + 1); // Temux documentation uses 1 relative
    DemoLogf("Link %s configured\n", 
      g_sTemuxInformation.bClockConfig[nLink] ? "is" : "is not");

    // OtxTemuxLinkSrcE
    DemoLogf("SONET/SDH 0x%X is ", g_sTemuxInformation.eLinkSrc[nLink]);
    switch(g_sTemuxInformation.eLinkSrc[nLink]) {
    case OTX_TEMUX_LINK_WORKING:
      DemoLogf("OTX_TEMUX_LINK_WORKING\n"); 
      break;

    case OTX_TEMUX_LINK_PROTECT:
      DemoLogf("OTX_TEMUX_LINK_PROTECT\n"); 
      break;

    default:
      DemoLogf("Not Configured OtxTemuxLinkSrcE\n"); 
      bDeviceConfigured = OTX_FALSE;
    }
    // OtxTemuxDataLinkSrcE
    DemoLogf("SONET/SDH 0x%X is ", g_sTemuxInformation.eDataLinkSource[nLink]);
    switch(g_sTemuxInformation.eDataLinkSource[nLink]) {
    case OTX_TEMUX_DATALINK_PROTECT:
      DemoLogf("OTX_TEMUX_DATALINK_PROTECT\n"); 
      break;

    case OTX_TEMUX_DATALINK_WORKING:
      DemoLogf("OTX_TEMUX_DATALINK_WORKING\n"); 
      break;

    default:
      DemoLogf("Not Configured OtxTemuxDataLinkSrcE\n"); 
      bDeviceConfigured = OTX_FALSE;
    }    
    // OtxTemuxSts12ConcatE
    DemoLogf("SONET/SDH 0x%X is ", g_sTemuxInformation.eSts12Concatenation[nLink]);
    switch(g_sTemuxInformation.eSts12Concatenation[nLink]) {
    case OTX_TEMUX_4_STS3:
      DemoLogf("OTX_TEMUX_4_STS3\n"); 
      break;

    case OTX_TEMUX_STS12:
      DemoLogf("OTX_TEMUX_STS12\n"); 
      break;

    default:
      DemoLogf("Not Configured OtxTemuxSts12ConcatE\n"); 
      bDeviceConfigured = OTX_FALSE;
    }
    // OtxTemuxSts3ConcatE
    DemoLogf("SONET/SDH 0x%X is ", g_sTemuxInformation.eSts3Concatenation[nLink]);
    switch(g_sTemuxInformation.eSts3Concatenation[nLink]) {
    case OTX_TEMUX_3_STS1:
      DemoLogf("OTX_TEMUX_3_STS1\n"); 
      break;

    case OTX_TEMUX_STS3:
      DemoLogf("OTX_TEMUX_STS3\n"); 
      break;

    default:
      DemoLogf("Not Configured OtxTemuxSts3ConcatE\n"); 
      bDeviceConfigured = OTX_FALSE;
    }
    // OtxTemuxClockSrcE
    DemoLogf("SONET/SDH 0x%X is ", g_sTemuxInformation.eClkSrcMode[nLink]);
    switch(g_sTemuxInformation.eClkSrcMode[nLink]) {
    case OTX_TEMUX_CLOCK_INTERNAL:
      DemoLogf("OTX_TEMUX_CLOCK_INTERNAL\n"); 
      break;

    case OTX_TEMUX_CLOCK_LOOP:
      DemoLogf("OTX_TEMUX_CLOCK_LOOP\n"); 
      break;

    default:
      DemoLogf("Not Configured OtxTemuxClockSrcE\n"); 
      bDeviceConfigured = OTX_FALSE;
    }

    DemoLogf("Source data to the destination port\n");
    eDestPort = g_sTemuxInformation.sStreamConnection[nLink].eDataLinkDestination;
    // Use -1 as not configured or used
    if(eDestPort == -1) {
      DemoLogf("Destination port link%d not configured\n", nLink);
    }
    else {
      OtxTemuxSrcPortE *pDataSrc;
      // OtxTemuxDestPortE
      DemoLogf("SONET/SDH 0x%X is ", eDestPort);
      switch(eDestPort) {
      case OTX_TEMUX_DEST_PORT_HOPP:
        DemoLogf("OTX_TEMUX_DEST_PORT_HOPP\n"); 
        break;

      case OTX_TEMUX_DEST_PORT_ESSI_W:
        DemoLogf("OTX_TEMUX_DEST_PORT_ESSI_W\n"); 
        break;

      case OTX_TEMUX_DEST_PORT_ESSI_P:
        DemoLogf("OTX_TEMUX_DEST_PORT_ESSI_P\n"); 
        break;

      case OTX_TEMUX_DEST_PORT_OC3_12:
        DemoLogf("OTX_TEMUX_DEST_PORT_OC3_12\n"); 
        break;

      default:
        DemoLogf("Not Configured OtxTemuxDestPortE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }

      pDataSrc = g_sTemuxInformation.sStreamConnection[nLink].eDataLinkSource;
      for(nSrcPort=0; nSrcPort < g_nLink; nSrcPort++) {
        // OtxTemuxSrcPortE
        DemoLogf("Destination port%d - 0x%X ", 
                 nSrcPort, 
                 pDataSrc[nSrcPort]);
        switch(pDataSrc[nSrcPort]) {
        case OTX_TEMUX_SRC_OC3_00:
          DemoLogf("OTX_TEMUX_SRC_OC3_00\n"); 
          break;

        case OTX_TEMUX_SRC_OC3_01:
          DemoLogf("OTX_TEMUX_SRC_OC3_01\n"); 
          break;

        case OTX_TEMUX_SRC_OC3_02:
          DemoLogf("OTX_TEMUX_SRC_OC3_02\n"); 
          break;

        case OTX_TEMUX_SRC_OC3_03:
          DemoLogf("OTX_TEMUX_SRC_OC3_03\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_10:
          DemoLogf("OTX_TEMUX_SRC_ESSI_10\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_11:
          DemoLogf("OTX_TEMUX_SRC_ESSI_11\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_12:
          DemoLogf("OTX_TEMUX_SRC_ESSI_12\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_13:
          DemoLogf("OTX_TEMUX_SRC_ESSI_13\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_20:
          DemoLogf("OTX_TEMUX_SRC_ESSI_20\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_21:
          DemoLogf("OTX_TEMUX_SRC_ESSI_21\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_22:
          DemoLogf("OTX_TEMUX_SRC_ESSI_22\n"); 
          break;

        case OTX_TEMUX_SRC_ESSI_23:
          DemoLogf("OTX_TEMUX_SRC_ESSI_23\n"); 
          break;

        case OTX_TEMUX_SRC_HOPP_30:
          DemoLogf("OTX_TEMUX_SRC_HOPP_30\n"); 
          break;

        case OTX_TEMUX_SRC_HOPP_31:
          DemoLogf("OTX_TEMUX_SRC_HOPP_31\n"); 
          break;

        case OTX_TEMUX_SRC_HOPP_32:
          DemoLogf("OTX_TEMUX_SRC_HOPP_32\n"); 
          break;

        case OTX_TEMUX_SRC_HOPP_33:
          DemoLogf("OTX_TEMUX_SRC_HOPP_33\n"); 
          break;

        default:
          DemoLogf("Destination port not Configured OtxTemuxSrcPortE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
      }// end of Source data to the destination port
    }

    DemoLogf("   PDH Subsystem configuration\n");

    for(nSts1=0; nSts1 < OTX_TEMUX_MAX_STS1; nSts1++) {
      DemoLogf("   PDH Subsystem %d\n", nSts1 + 1); // Temux documentation uses 1 relative
      // OtxTemuxSysIfModeE
      DemoLogf("   PDH Subsystem  0x%X - ", 
        g_sTemuxInformation.eSysIfMode[nLink][nSts1]); 
      switch(g_sTemuxInformation.eSysIfMode[nLink][nSts1]) {
      case OTX_TEMUX_SBI:
        DemoLogf("OTX_TEMUX_SBI\n"); 
        break;

      case OTX_TEMUX_SBITR:
        DemoLogf("OTX_TEMUX_SBITR\n"); 
        break;

      case OTX_TEMUX_HMVIP:
        DemoLogf("OTX_TEMUX_HMVIP\n"); 
        break;

      default:
        DemoLogf("   Not Configured OtxTemuxSysIfModeE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }
      // OtxTemuxSpeDestE
      DemoLogf("   PDH Subsystem  0x%X - ", 
        g_sTemuxInformation.eSpeDest[nLink][nSts1]); 
      switch(g_sTemuxInformation.eSpeDest[nLink][nSts1]) {
      case OTX_TEMUX_TRIBXC:
        DemoLogf("OTX_TEMUX_TRIBXC\n"); 
        break;

      case OTX_TEMUX_MAPPER:
        DemoLogf("OTX_TEMUX_MAPPER\n"); 
        break;

      default:
        DemoLogf("   Not Configured OtxTemuxSpeDestE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }
      // OtxTemuxPhdTribE
      DemoLogf("   PDH Subsystem  0x%X - ", 
        g_sTemuxInformation.ePhdTribType[nLink][nSts1]); 
      switch(g_sTemuxInformation.ePhdTribType[nLink][nSts1]) {
      case OTX_TEMUX_T1:
        DemoLogf("OTX_TEMUX_T1\n"); 
        break;

      case OTX_TEMUX_E1:
        DemoLogf("OTX_TEMUX_E1\n"); 
        break;

      default:
        DemoLogf("   Not Configured OtxTemuxPhdTribE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }
      // OtxTemuxMapSpeSrcE
      DemoLogf("   PDH Subsystem  0x%X - ", 
        g_sTemuxInformation.eMapSpeSrc[nLink][nSts1]); 
      switch(g_sTemuxInformation.eMapSpeSrc[nLink][nSts1]) {
      case OTX_TEMUX_T1E1:
        DemoLogf("OTX_TEMUX_T1E1\n"); 
        break;

      case OTX_TEMUX_DS3E3:
        DemoLogf("OTX_TEMUX_DS3E3\n"); 
        break;

      default:
        DemoLogf("   Not Configured OtxTemuxMapSpeSrcE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }
      // OtxTemuxSts1PayloadE
      DemoLogf("   PDH Subsystem  0x%X - ", 
        g_sTemuxInformation.eSts1Payload[nLink][nSts1]); 
      switch(g_sTemuxInformation.eSts1Payload[nLink][nSts1]) {
      case OTX_TEMUX_STS_1:
        DemoLogf("OTX_TEMUX_STS_1\n"); 
        break;

      case OTX_TEMUX_7_VT_GROUP:
        DemoLogf("OTX_TEMUX_7_VT_GROUP\n"); 
        break;

      default:
        DemoLogf("Not Configured OtxTemuxSts1PayloadE\n"); 
        bDeviceConfigured = OTX_FALSE;
      }

      DemoLogf("   PDH Subsystem: Virtual Tributaries (VT)\n"); 
      for(nVtGroup=0; nVtGroup < OTX_TEMUX_MAX_VTGROUP; nVtGroup++) {
        // OtxTemuxVtTypeE
        DemoLogf("   VT Group%d   0x%X - ",
          nVtGroup + 1, // Temux documentaion uses 1 relative
          g_sTemuxInformation.eVtGroupPayload[nLink][nSts1][nVtGroup]); 
        switch(g_sTemuxInformation.eVtGroupPayload[nLink][nSts1][nVtGroup]) {
        case OTX_TEMUX_VT_2:
          DemoLogf("OTX_TEMUX_VT_2\n"); 
          break;

        case OTX_TEMUX_VT_15:
          DemoLogf("OTX_TEMUX_VT_15\n"); 
          break;

        default:
          DemoLogf("   Not Configured OtxTemuxVtTypeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
      }//for nVtGroup

      DemoLogf("   PDH Subsystem: Tributaries \n"); 
      for(nTrib=0; nTrib < OTX_TEMUX_MAX_TRIB; nTrib++) {
        // Tributary
        // OtxTemuxFrameModeE
        DemoLogf("      Tributaries %d \n", nTrib + 1); // Temux API uses 1 relative
        // OTX_BOOL

        DemoLogf("      0x%X - ",
          g_sTemuxInformation.eFrameMode[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eFrameMode[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_T1_UNFRAMED:
          DemoLogf("OTX_TEMUX_T1_UNFRAMED\n"); 
          break;

        case OTX_TEMUX_T1_ESF:
          DemoLogf("OTX_TEMUX_T1_ESF\n"); 
          break;

        case OTX_TEMUX_T1_SF:
          DemoLogf("OTX_TEMUX_T1_SF\n"); 
          break;

        case OTX_TEMUX_E1_UNFRAMED:
          DemoLogf("OTX_TEMUX_E1_UNFRAMED\n"); 
          break;

        case OTX_TEMUX_E1_STANDARD:
          DemoLogf("OTX_TEMUX_E1_STANDARD\n"); 
          break;

        case OTX_TEMUX_E1_CRC:
          DemoLogf("OTX_TEMUX_E1_CRC\n"); 
          break;

        case OTX_TEMUX_E1_CAS:
          DemoLogf("OTX_TEMUX_E1_CAS\n"); 
          break;

        case OTX_TEMUX_E1_CAS_CRC:
          DemoLogf("OTX_TEMUX_E1_CAS_CRC\n"); 
          break;

        default:
          DemoLogf("   Not Configured OtxTemuxFrameModeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxELSTE
        DemoLogf("      0x%X - ",
          g_sTemuxInformation.eBypassModeIngress[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eBypassModeIngress[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_ELST_DATAPATH:
          DemoLogf("OTX_TEMUX_ELST_DATAPATH\n"); 
          break;

        case OTX_TEMUX_ELST_BYPASSED:
          DemoLogf("OTX_TEMUX_ELST_BYPASSED\n"); 
          break;

        default:
          DemoLogf("Not Configured OtxTemuxELSTE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxRxTimingE
        DemoLogf("      0x%X - ",
                 g_sTemuxInformation.eTimingIngress[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eTimingIngress[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_RX_TT_RTDM:
          DemoLogf("OTX_TEMUX_RX_TT_RTDM\n"); 
          break;

        case OTX_TEMUX_RX_TT_EXSBI:
          DemoLogf("OTX_TEMUX_RX_TT_EXSBI\n"); 
          break;

        case OTX_TEMUX_RX_TT_MX13:
          DemoLogf("OTX_TEMUX_RX_TT_MX13\n"); 
          break;

        case OTX_TEMUX_RX_TT_MUX:
          DemoLogf("OTX_TEMUX_RX_TT_MUX\n"); 
          break;

        default:
          DemoLogf("   Not ConfiguredOtxTemuxRxTimingE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxELSTE
        DemoLogf("      0x%X - ",
                 g_sTemuxInformation.eBypassModeEgress[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eBypassModeEgress[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_ELST_DATAPATH:
          DemoLogf("OTX_TEMUX_ELST_DATAPATH\n"); 
          break;

        case OTX_TEMUX_ELST_BYPASSED:
          DemoLogf("OTX_TEMUX_ELST_BYPASSED\n"); 
          break;

        default:
          DemoLogf("   Not Configured OtxTemuxELSTE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxTxTimingE
        DemoLogf("      0x%X - ",
                 g_sTemuxInformation.eTimingEgress[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eTimingEgress[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_TX_TT_LOOP:
          DemoLogf("OTX_TEMUX_TX_TT_LOOP\n"); 
          break;

        case OTX_TEMUX_TX_TT_CTCLK:
          DemoLogf("OTX_TEMUX_TX_TT_CTCLK\n"); 
          break;

        case OTX_TEMUX_TX_TT_DSSI:
          DemoLogf("OTX_TEMUX_TX_TT_DSSI\n"); 
          break;

        case OTX_TEMUX_TX_TT_TISS:
          DemoLogf("OTX_TEMUX_TX_TT_TISS\n"); 
          break;

        case OTX_TEMUX_TX_TT_MUX:
          DemoLogf("OTX_TEMUX_TX_TT_MUX\n"); 
          break;

        default:
          DemoLogf("   Not Configured OtxTemuxTxTimingE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxMapTribModeE
        DemoLogf("      0x%X - ",
                 g_sTemuxInformation.eSrcOrDestEgress[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eSrcOrDestEgress[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_BIT_ASYN:
          DemoLogf("OTX_TEMUX_BIT_ASYN\n"); 
          break;

        case OTX_TEMUX_BYTE_SYNC:
          DemoLogf("OTX_TEMUX_BYTE_SYNC\n"); 
          break;

        case OTX_TEMUX_TVT:
          DemoLogf("OTX_TEMUX_TVT\n"); 
          break;

        default:
          DemoLogf("   Not Configured OtxTemuxMapTribModeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxMapTribModeE
        DemoLogf("      Mapped 0x%X - ",
                 g_sTemuxInformation.eMapTribMode[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eMapTribMode[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_BIT_ASYN:
          DemoLogf("OTX_TEMUX_BIT_ASYN\n"); 
          break;

        case OTX_TEMUX_BYTE_SYNC:
          DemoLogf("OTX_TEMUX_BYTE_SYNC\n"); 
          break;

        case OTX_TEMUX_TVT:
          DemoLogf("OTX_TEMUX_TVT\n"); 
          break;

        default:
          DemoLogf("Not Configured OtxTemuxMapTribModeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxMapTribModeE
        DemoLogf("      Mapped 0x%X - ",
                 g_sTemuxInformation.eMapPhdTribType[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eMapPhdTribType[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_T1:
          DemoLogf("OTX_TEMUX_T1\n"); 
          break;

        case OTX_TEMUX_E1:
          DemoLogf("OTX_TEMUX_E1\n"); 
          break;

        default:
          DemoLogf("Not Configured OtxTemuxMapTribModeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }
        // OtxTemuxVtTypeE
        DemoLogf("      Mapped 0x%X - ",
                 g_sTemuxInformation.eMapSonetTribType[nLink][nSts1][nTrib]); 
        switch(g_sTemuxInformation.eMapSonetTribType[nLink][nSts1][nTrib]) {
        case OTX_TEMUX_VT_2:
          DemoLogf("OTX_TEMUX_VT_2\n"); 
          break;

        case OTX_TEMUX_VT_15:
          DemoLogf("OTX_TEMUX_VT_15\n"); 
          break;

        default:
          DemoLogf("Not Configured OtxTemuxVtTypeE\n"); 
          bDeviceConfigured = OTX_FALSE;
        }

        DemoLogf("      T1E1 SBI bus %s enabled\n", 
          g_sTemuxInformation.bT1E1Enable[nLink][nSts1][nTrib] ? "is" : "is not");
        // If it is not enabled then no need to display configuration
        if(g_sTemuxInformation.bT1E1Enable[nLink][nSts1][nTrib]) {
          // OtxTemuxSyncE
          DemoLogf("      0x%X - ",
                   g_sTemuxInformation.eT1E1SyncTrib[nLink][nSts1][nTrib]); 
          switch(g_sTemuxInformation.eT1E1SyncTrib[nLink][nSts1][nTrib]) {
          case OTX_TEMUX_ASYNC:
            DemoLogf("OTX_TEMUX_ASYNC\n"); 
            break;

          case OTX_TEMUX_SYNC:
            DemoLogf("OTX_TEMUX_SYNC\n"); 
            break;

          default:
            DemoLogf("   Not Configured OtxTemuxSyncE\n"); 
            bDeviceConfigured = OTX_FALSE;
          }

          // OtxTemuxModeE
          DemoLogf("      0x%X - ",
                   g_sTemuxInformation.eT1E1RateMode[nLink][nSts1][nTrib]); 
          switch(g_sTemuxInformation.eT1E1RateMode[nLink][nSts1][nTrib]) {
          case OTX_TEMUX_SBI_T1E1:
            DemoLogf("OTX_TEMUX_SBI_T1E1\n"); 
            break;

          case OTX_TEMUX_SBI_TVT:
            DemoLogf("OTX_TEMUX_SBI_TVT\n"); 
            break;

          default:
            DemoLogf("   Not Configured OtxTemuxModeE\n"); 
            bDeviceConfigured = OTX_FALSE;
          }
        } // if g_sTemuxInformation.bT1E1Enable
      }//for nTrib
    }//for nSts1
  }//for nLink

  if(bDeviceConfigured == OTX_FALSE) {
    // No need to check other settings the system is not configured
    DemoLogf("Temux system configuration is not completely configured\n");
  }

  return;
}


/*******************************************************************************
  @func
    Configure the Temux device

                  4 x STM-1 Working Links - loop timed
                  Each STM-1 has the following mapping:
                     STM-1 <-> AUG-1 <-> 1 x AU-4
                  Each AU-4 has the following mapping:
                     AU-4<-> 3 x TUG-3
                  Each TUG-3 has the following mapping:
                     TUG-3 <-> 7 x TUG-2
                  Each TUG-2 has the following mapping:
                     TUG-2 <-> 3 x TU-12
                  Each TU-12 contains a bit asynchronously mapped E1
                  Each E1 has Standard framing format.
                  The SBI System Interface is not supported on this board
                  For HMVIP System Interface, E1 tributaries are internally 
                  timed to CTCLK.  

*******************************************************************************/
void DemoTemuxInterfaceUserDefinedE1(void)
{
  OTX_UINT16            nTrib;
  OTX_UINT16            nTug2;
  OTX_UINT16            nTug3;
  OTX_UINT16            nSpe;
  OTX_UINT16            nSlice;
  OTX_UINT16            nInterfaceSelect;
  OtxTemuxSts12ConcatE  eSts12Concatenation;
  OtxTemuxSts3ConcatE   eSts3Concatenation;
  OtxTemuxIntfModeE     eIntfMode;
  OtxTemuxClockSrcE     eClkSrcMode;
  OtxTemuxBlockModeE    eBlockMode;
  OtxTemuxDataLinkSrcE  eDataLinkSource;
  OtxTemuxMapSpeSrcE    eSpeSource;
  OtxTemuxSpeDestE      eSpeDest;
  OtxTemuxPhdTribE      eSpePayload; 
  OtxTemuxSts3PayloadE  eSts3Payload;
  OtxTemuxSts3cPayloadE eSts3cPayload;
  OtxTemuxSts12PayloadE eSts12Payload;
  OtxTemuxLinkSrcE      eProtection;
  OtxTemuxRxTimingE     eRxTimingMode;
  OtxTemuxTxTimingE     eTxTimingMode;
  OtxTemuxMapTribModeE  eLinkSrcOrDest;
  OtxTemuxELSTE         eBypass; 
  OtxTemuxVtTypeE       eSonetTribType;
  OtxTemuxPhdTribE      ePdhTribType;
  OtxTemuxFrameModeE    eFrameMode;
  OtxTemuxTug3PayloadE eTug3Payload;
  OtxTemuxTug2PayloadE eTug2Payload;
  OtxTemuxSysIfModeE    eSysIfMode;
  OtxTemuxSrcPortE      eSourcePort[4];
  OTX_BOOL              bEnable = 1;
  OTX_BOOL              bHmvip_tfall, bHmvip_rfall, bUnused;
  OTX_RESULT            otxRc;

  // interface - 0 SBI interface, 1 HMVIP interface
  nInterfaceSelect = OTX_TEMUX_INTERFACE_HMVIP;

  //************************************************************************//
  // Device reset with CSU intialization
  //************************************************************************//
  // initialize the CSU as part of device initialization.
  otxRc = OtxTemuxInitialize(g_hTemux);
  DemoCheckDevRc(g_hTemux, otxRc);
      
  //************************************************************************//
  // SONET subsystem configuration 
  //************************************************************************//

  //Functions from sonet subsystem 
  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Initialize the DCRU_622 and JAT_622_E for all link
    otxRc = OtxTemuxSonetClockCfg(g_hTemux, nSlice);
    DemoCheckDevRc(g_hTemux, otxRc);
  }

  // Configure the device to use 4 x 155 Mbps links instead of 1 x 622 Mbps
  // links on the physical interface.
  eIntfMode = OTX_TEMUX_INTF_MODE_OC3_STM1; // up to 4 x OC-3/STM-1
  otxRc = OtxTemuxSonetIntfCfg(g_hTemux, eIntfMode);
  DemoCheckDevRc(g_hTemux, otxRc);


  eClkSrcMode = g_demoEnv.eClkSrcMode;  
  DemoLogf("Sonet Timing %s\n",
           eClkSrcMode ? "OTX_TEMUX_CLOCK_LOOP" : "OTX_TEMUX_CLOCK_INTERNAL");

  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Configure the working serial link for use.
    eProtection = OTX_TEMUX_LINK_WORKING; // use working link as source
    otxRc = OtxTemuxSonetLinkCfg(g_hTemux, eProtection, nSlice, bEnable);
    DemoCheckDevRc(g_hTemux, otxRc);

    // The OtxTemuxSonetPayloadCfg has two usage cases. The first case is when 
    // the SONET/SDH line interface is configured for one OC-12/STM-4 physical 
    // link that contains STS-12c payload. The second case is when the SONET/SDH 
    // line interface is configured for up to 4 x OC-3/STM-1 physical links or 
    // an OC-12/STM-4 that is configured for 4 x STS-3.

    // Usage case 1 - one OC-12/STM-4 physical link with STS-12c payload:
    // The function is called with link set to 1, sts12Concatenation set to 1 
    // and the value of sts3concatenation is dont care.

    // Usage case 2 - up to 4 x OC-3/STM-1 physical links or 1 x OC-12/STM-4 
    // with 4 x STS-3 Payload:

    // When 4 x STS-3 payload is used, sts12Concatenation must always be set to 0. 
    // For each link (1 to 4), the sts3Concatenation input parameter will specify 
    // whether the respective STS-3/AUG-1 contains concatenated (STS-3c/AU-4) 
    // or non-concatenated (3 x STS-1/AU-3) payload.

    // Configure each STS-3 for
    eSts12Concatenation = OTX_TEMUX_4_STS3; // 4 x STS-3 instead of 1 x STS-12C
    eSts3Concatenation  = OTX_TEMUX_STS3;   // 1 x STS-3c/AU4 not 3 x STS-1/AU3
    otxRc = OtxTemuxSonetPayloadCfg(g_hTemux, 
                                    nSlice, 
                                    eSts12Concatenation, 
                                    eSts3Concatenation);
    DemoCheckDevRc(g_hTemux, otxRc); 
   
    // Configure the SONET timing to be internally timed or loop timed
    otxRc = OtxTemuxSonetTimingCfg(g_hTemux, nSlice, eClkSrcMode);
    DemoCheckDevRc(g_hTemux, otxRc);

    // Configure the serial multiplexers to provide data to the APS
    // from the working link instead of the protect link.
    eDataLinkSource = OTX_TEMUX_DATALINK_WORKING; // use working link as source
    otxRc = OtxTemuxSonetDataPathCfg(g_hTemux, nSlice, eDataLinkSource);
    DemoCheckDevRc(g_hTemux, otxRc);
  }

  // Configure Rx SONET Links 1 to 4 as the source for HOPP Receive
  eSourcePort[0] = OTX_TEMUX_SRC_OC3_00;
  eSourcePort[1] = OTX_TEMUX_SRC_OC3_01;
  eSourcePort[2] = OTX_TEMUX_SRC_OC3_02;
  eSourcePort[3] = OTX_TEMUX_SRC_OC3_03;
  otxRc = OtxTemuxStsiDestPortCfg(g_hTemux, 
                                  OTX_TEMUX_DEST_PORT_HOPP, 
                                  eSourcePort);
  DemoCheckDevRc(g_hTemux, otxRc);
  // Configure HOPP Transmit as the source for Tx SONET Links 1 to 4
  eSourcePort[0] = OTX_TEMUX_SRC_HOPP_30;
  eSourcePort[1] = OTX_TEMUX_SRC_HOPP_31;
  eSourcePort[2] = OTX_TEMUX_SRC_HOPP_32;
  eSourcePort[3] = OTX_TEMUX_SRC_HOPP_33;
  otxRc = OtxTemuxStsiDestPortCfg(g_hTemux, 
                                  OTX_TEMUX_DEST_PORT_OC3_12, 
                                  eSourcePort);
  DemoCheckDevRc(g_hTemux, otxRc);
  
  // Configure RMSU for BYPASS mode.
  eBlockMode = OTX_TEMUX_MSU_MODE_BYPASS; // setting for BYPASS mode
  otxRc = OtxTemuxMsuCfg(g_hTemux, 
                         OTX_TEMUX_MSU_TYPE_RMSU, 
                         eBlockMode);
  DemoCheckDevRc(g_hTemux, otxRc);
  // Configure TMSU for Bypass mode.
  otxRc = OtxTemuxMsuCfg(g_hTemux, 
                         OTX_TEMUX_MSU_TYPE_TMSU, 
                         eBlockMode);
  DemoCheckDevRc(g_hTemux, otxRc);

  // Perform global config of HOPP, LOPP and LOPP_TX
  otxRc = OtxTemuxHoLoPtrCfg(g_hTemux, 20);
  DemoCheckDevRc(g_hTemux, otxRc);

  // Configure STS12 payload for 4 x STS-3
  eSts12Payload = OTX_TEMUX_4_STS_3; // setting for up to 4 x STS-3
  otxRc = OtxTemuxSts12PayloadCfg(g_hTemux, eSts12Payload);
  DemoCheckDevRc(g_hTemux, otxRc);

  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Configure STS-3 for STS-3c/AU-4 payload
    eSts3Payload = OTX_TEMUX_STS_3; // setting for STS-3c/AU-4
    otxRc = OtxTemuxSts3PayloadCfg(g_hTemux, nSlice, eSts3Payload);
    DemoCheckDevRc(g_hTemux, otxRc);

    // Configure STS-3c for 3 x TUG-3 payload
    eSts3cPayload = OTX_TEMUX_3C_STS_1; // setting for 3 x TUG-3
    otxRc = OtxTemuxSts3cPayloadCfg(g_hTemux, nSlice, eSts3cPayload);
    DemoCheckDevRc(g_hTemux, otxRc);

    for (nTug3 = 1; nTug3 <= OTX_TEMUX_MAX_STS1; nTug3++) {
      // Configure each TUG-3 for TUG-2 payload
      eTug3Payload = OTX_TEMUX_7_TUG_2; // setting for 7 x TUG-2
      otxRc = OtxTemuxTug3PayloadCfg(g_hTemux, 
                                     nSlice, 
                                     nTug3, 
                                     eTug3Payload);
      DemoCheckDevRc(g_hTemux, otxRc);

      for (nTug2 = 1; nTug2 <= OTX_TEMUX_MAX_VTGROUP; nTug2++) {
          // configure each TUG-2 for TU-12 payload
          eTug2Payload = OTX_TEMUX_TU_12; // setting for TU-12
          otxRc = OtxTemuxTug2PayloadCfg(g_hTemux, 
                                         nSlice, 
                                         nTug3, 
                                         nTug2, 
                                         eTug2Payload);
          DemoCheckDevRc(g_hTemux, otxRc);
      }
    }
  }

  //************************************************************************//    
  // PDH subsystem configuration
  //************************************************************************//

  // OTX_TEMUX_INTERFACE_HMVIP is supported
  // configure the system side interface for HMVIP
  // specify HMVIP mode of operation for system side interface.
  eSysIfMode = OTX_TEMUX_HMVIP;   
                  
  bHmvip_tfall = 0; // HMVIP_TFALL
  bHmvip_rfall = 0; // HMVIP_RFALL
  bUnused      = 0; // unused set to 0
                          
  otxRc = OtxTemuxSysIfModeCfg(g_hTemux, 
                               eSysIfMode, 
                               bUnused, 
                               bHmvip_tfall, 
                               bHmvip_rfall, 
                               bUnused, 
                               bUnused);  
  DemoCheckDevRc(g_hTemux, otxRc); 

  for(nSlice = 1; nSlice <= g_nLink; nSlice++) {
    for(nSpe = 1; nSpe <= OTX_TEMUX_MAX_STS1; nSpe++) {  
      // configure each slice and spe for E1 mode
      eSpeDest    = OTX_TEMUX_MAPPER; // E1 tribs associated with Mapper not MX13
      eSpePayload = OTX_TEMUX_E1;     // E1 tribs are specified not T1
      otxRc = OtxTemuxE1T1SpeCfg(g_hTemux, 
                                 nSlice, 
                                 nSpe, 
                                 eSysIfMode, 
                                 eSpeDest, 
                                 eSpePayload);
      DemoCheckDevRc(g_hTemux, otxRc); 
      // configure each slice and spe mapper to process data from the
      // T1E1 subsystem instead of the DS3E3 subsystem
      eSpeSource = OTX_TEMUX_T1E1;
      otxRc = OtxTemuxMapperSpeCfg(g_hTemux, nSlice, nSpe, eSpeSource);
      DemoCheckDevRc(g_hTemux, otxRc); 

      for(nTrib = 1; nTrib <= OTX_TEMUX_MAX_TRIB; nTrib++) {
        // configure each slice, spe and trib for E1
        eFrameMode = g_demoEnv.eE1FramingMode; // E1 framing
        // NOTE: UNFRAMED is not supported by the HMVIP interface
        otxRc = OtxTemuxE1T1TribFrameCfg(g_hTemux, 
                                         nSlice, 
                                         nSpe, 
                                         nTrib, 
                                         eFrameMode);
        DemoCheckDevRc(g_hTemux, otxRc); 
        // configure each slice, spe and trib ingress timing
        eBypass = OTX_TEMUX_ELST_DATAPATH; // ELST must be part of data path
        eRxTimingMode = OTX_TEMUX_RX_TT_RTDM; // Data source is from the RTDM
        otxRc = OtxTemuxE1T1IngressTribTimingCfg(g_hTemux, 
                                                 nSlice, 
                                                 nSpe, 
                                                 nTrib, 
                                                 eBypass, 
                                                 eRxTimingMode);
        DemoCheckDevRc(g_hTemux, otxRc);
        // configure each slice, spe and trib egress timing
        eBypass = OTX_TEMUX_ELST_DATAPATH; // ELST must be part of data path
        eTxTimingMode = OTX_TEMUX_TX_TT_CTCLK; // CTCLK Internal Timing
        // trib destined for bit async mapping
        eLinkSrcOrDest = g_demoEnv.eLinkSrcOrDest; //OTX_TEMUX_BIT_ASYN;  
        otxRc = OtxTemuxE1T1EgressTribTimingCfg(g_hTemux, 
                                                nSlice, 
                                                nSpe, 
                                                nTrib, 
                                                eBypass, 
                                                eTxTimingMode,
                                                eLinkSrcOrDest);
        DemoCheckDevRc(g_hTemux, otxRc);
        // configure each slice, spe and trib for bit async mapping
        eLinkSrcOrDest = g_demoEnv.eLinkSrcOrDest; //OTX_TEMUX_BIT_ASYN; // bit asynchronous mapping
        ePdhTribType   = OTX_TEMUX_E1;       // E1 tributary
        eSonetTribType = OTX_TEMUX_VT_2;     // VT2 VT Type
        otxRc = OtxTemuxMapperTribCfg(g_hTemux, 
                                      nSlice, 
                                      nSpe, 
                                      nTrib, 
                                      eLinkSrcOrDest, 
                                      ePdhTribType, 
                                      eSonetTribType);
        DemoCheckDevRc(g_hTemux, otxRc);

      } // For(nTrib)
    } // For(nSpe)
  } // For(nSlice)
}


/*******************************************************************************
  @func
    Configure the Temux device

                  4 x OC-3 Working Links - Internally timed
                  Each OC-3 has the following mapping:
                     OC-3 <-> STS-3 <-> 3 x STS-1
                  Each STS-1 has the following mapping:
                     STS-1 <-> 7 x VT-Group
                  Each VT-Group has the following mapping:
                     VT-Group <-> 4 x VT-1.5
                  Each VT-1.5 contains a bit asynchronously mapped T1
                  Each T1 has ESF framing format.
                  For SBI System Interface is not supported on this board
                  For HMVIP System Interface, T1 tributaries are internally 
                  timed to CTCLK.  

*******************************************************************************/
void DemoTemuxInterfaceUserDefinedT1(void)
{
  OTX_UINT16            nTrib;
  OTX_UINT16            nSts1;
  OTX_UINT16            nSpe;
  OTX_UINT16            nSlice;
  OTX_UINT16            nVtGroup;
  OTX_UINT16            nInterfaceSelect;
  OtxTemuxSts12ConcatE  eSts12Concatenation;
  OtxTemuxSts3ConcatE   eSts3Concatenation;
  OtxTemuxIntfModeE     eIntfMode;
  OtxTemuxClockSrcE     eClkSrcMode;
  OtxTemuxBlockModeE    eBlockMode;
  OtxTemuxDataLinkSrcE  eDataLinkSource;
  OtxTemuxMapSpeSrcE    eSpeSource;
  OtxTemuxSpeDestE      eSpeDest;
  OtxTemuxPhdTribE      eSpePayload; 
  OtxTemuxSts1PayloadE  eSts1Payload;
  OtxTemuxSts3PayloadE  eSts3Payload;
  OtxTemuxSts12PayloadE eSts12Payload;
  OtxTemuxLinkSrcE      eProtection;
  OtxTemuxRxTimingE     eRxTimingMode;
  OtxTemuxTxTimingE     eTxTimingMode;
  OtxTemuxMapTribModeE  eLinkSrcOrDest;
  OtxTemuxELSTE         eBypass; 
  OtxTemuxVtTypeE       eSonetTribType;
  OtxTemuxPhdTribE      ePdhTribType;
  OtxTemuxFrameModeE    eFrameMode;
  OtxTemuxVtTypeE       eVtGroupPayload;
  OtxTemuxSysIfModeE    eSysIfMode;
  OtxTemuxSrcPortE      eSourcePort[4];
  OTX_BOOL              bEnable = 1;
  OTX_BOOL              bHmvip_tfall, bHmvip_rfall, bUnused;
  OTX_RESULT            otxRc;

  // interface - 0 SBI interface, 1 HMVIP interface
  // HMVIP interface is supported by Godar board
  nInterfaceSelect = OTX_TEMUX_INTERFACE_HMVIP; 

  //************************************************************************//
  // Device reset with CSU intialization
  //************************************************************************//
  // initialize the CSU as part of device initialization.
  otxRc = OtxTemuxInitialize(g_hTemux);
  DemoCheckDevRc(g_hTemux, otxRc);
      
  //************************************************************************//
  // SONET subsystem configuration 
  //************************************************************************//

  //Functions from sonet subsystem 
  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Initialize the DCRU_622 and JAT_622_E for all link
    otxRc = OtxTemuxSonetClockCfg(g_hTemux, nSlice);
    DemoCheckDevRc(g_hTemux, otxRc);
  }

  // Configure the device to use 4 x 155 Mbps links instead of 1 x 622 Mbps
  // links on the physical interface.
  eIntfMode = OTX_TEMUX_INTF_MODE_OC3_STM1; // up to 4 x OC-3/STM-1
  otxRc = OtxTemuxSonetIntfCfg(g_hTemux, eIntfMode);
  DemoCheckDevRc(g_hTemux, otxRc);


  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Configure the working serial link for use.
    eProtection = OTX_TEMUX_LINK_WORKING; // use working link as source
    otxRc = OtxTemuxSonetLinkCfg(g_hTemux, eProtection, nSlice, bEnable);
    DemoCheckDevRc(g_hTemux, otxRc);
    eProtection = OTX_TEMUX_LINK_PROTECT; // add protect link config
    otxRc = OtxTemuxSonetLinkCfg(g_hTemux, eProtection, nSlice, bEnable);
    DemoCheckDevRc(g_hTemux, otxRc);

    // The OtxTemuxSonetPayloadCfg has two usage cases. The first case is when 
    // the SONET/SDH line interface is configured for one OC-12/STM-4 physical 
    // link that contains STS-12c payload. The second case is when the SONET/SDH 
    // line interface is configured for up to 4 x OC-3/STM-1 physical links or 
    // an OC-12/STM-4 that is configured for 4 x STS-3.

    // Usage case 1 - one OC-12/STM-4 physical link with STS-12c payload:
    // The function is called with link set to 1, sts12Concatenation set to 1 
    // and the value of sts3concatenation is dont care.

    // Usage case 2 - up to 4 x OC-3/STM-1 physical links or 1 x OC-12/STM-4 
    // with 4 x STS-3 Payload:

    // When 4 x STS-3 payload is used, sts12Concatenation must always be set to 0. 
    // For each link (1 to 4), the sts3Concatenation input parameter will specify 
    // whether the respective STS-3/AUG-1 contains concatenated (STS-3c/AU-4) 
    // or non-concatenated (3 x STS-1/AU-3) payload.

    // Configure each STS-3 for
    eSts12Concatenation = OTX_TEMUX_4_STS3; // 4 x STS-3 instead of 1 x STS-12C
    eSts3Concatenation  = OTX_TEMUX_3_STS1; // 3 x STS-1 instead of 1 x STS-3c
    otxRc = OtxTemuxSonetPayloadCfg(g_hTemux, 
                                    nSlice, 
                                    eSts12Concatenation, 
                                    eSts3Concatenation);
    DemoCheckDevRc(g_hTemux, otxRc); 
   
    // Configure the SONET timing to be internally timed instead of
    // loop timed.
    DemoLogf("OTX_TEMUX_CLOCK_LOOP %d\n", nSlice); 
    eClkSrcMode = OTX_TEMUX_CLOCK_LOOP;// OTX_TEMUX_CLOCK_INTERNAL; // internal timing
    otxRc = OtxTemuxSonetTimingCfg(g_hTemux, nSlice, eClkSrcMode);
    DemoCheckDevRc(g_hTemux, otxRc);

    // Configure the serial multiplexers to provide data to the APS
    // from the working link instead of the protect link.
    if(nSlice == 0) {
      DemoLogf("Ingress data source protect\n"); 
      eDataLinkSource = OTX_TEMUX_DATALINK_PROTECT; // use protect link as source
    }
    else {
      DemoLogf("Ingress data source working\n"); 
      eDataLinkSource = OTX_TEMUX_DATALINK_WORKING; // use working link as source
    }
    otxRc = OtxTemuxSonetDataPathCfg(g_hTemux, nSlice, eDataLinkSource);
    DemoCheckDevRc(g_hTemux, otxRc);
  }

  // Configure Rx SONET Links 1 to 4 as the source for HOPP Receive
  DemoLogf("OTX_TEMUX_DEST_PORT_HOPP Rx OC-3 #1/Rx OC-12 STS-3 #1\n"); 
  eSourcePort[0] = OTX_TEMUX_SRC_OC3_00;
  eSourcePort[1] = OTX_TEMUX_SRC_OC3_01;
  eSourcePort[2] = OTX_TEMUX_SRC_OC3_02;
  eSourcePort[3] = OTX_TEMUX_SRC_OC3_03;
  otxRc = OtxTemuxStsiDestPortCfg(g_hTemux, 
                                  OTX_TEMUX_DEST_PORT_HOPP, 
                                  eSourcePort);
  DemoCheckDevRc(g_hTemux, otxRc);
  // Configure HOPP Transmit as the source for Tx SONET Links 1 to 4
  eSourcePort[0] = OTX_TEMUX_SRC_HOPP_30;
  eSourcePort[1] = OTX_TEMUX_SRC_HOPP_31;
  eSourcePort[2] = OTX_TEMUX_SRC_HOPP_32;
  eSourcePort[3] = OTX_TEMUX_SRC_HOPP_33;
  otxRc = OtxTemuxStsiDestPortCfg(g_hTemux, 
                                  OTX_TEMUX_DEST_PORT_OC3_12, 
                                  eSourcePort);
  DemoCheckDevRc(g_hTemux, otxRc);
  
  // Configure RMSU for BYPASS mode.
  eBlockMode = OTX_TEMUX_MSU_MODE_BYPASS; // setting for BYPASS mode
  otxRc = OtxTemuxMsuCfg(g_hTemux, 
                         OTX_TEMUX_MSU_TYPE_RMSU, 
                         eBlockMode);
  DemoCheckDevRc(g_hTemux, otxRc);
  // Configure TMSU for Bypass mode.
  otxRc = OtxTemuxMsuCfg(g_hTemux, 
                         OTX_TEMUX_MSU_TYPE_TMSU, 
                         eBlockMode);
  DemoCheckDevRc(g_hTemux, otxRc);

  // Perform global config of HOPP, LOPP and LOPP_TX
  otxRc = OtxTemuxHoLoPtrCfg(g_hTemux, 20); // Timeout of 20
  DemoCheckDevRc(g_hTemux, otxRc);

  // Configure STS12 payload for 4 x STS-3
  eSts12Payload = OTX_TEMUX_4_STS_3; // setting for up to 4 x STS-3
  otxRc = OtxTemuxSts12PayloadCfg(g_hTemux, eSts12Payload);
  DemoCheckDevRc(g_hTemux, otxRc);

     
  for (nSlice = 1; nSlice <= g_nLink; nSlice++) {
    // Configure each STS-3 for 3 x STS-1 payload
    eSts3Payload = OTX_TEMUX_3_STS_1; // setting for 3 x STS-1/AU-3
    otxRc = OtxTemuxSts3PayloadCfg(g_hTemux, nSlice, eSts3Payload);
    DemoCheckDevRc(g_hTemux, otxRc);

    for (nSts1 = 1; nSts1 <= OTX_TEMUX_MAX_STS1; nSts1++) {
      // Configure each STS-1 for VT-Group payload
      eSts1Payload = OTX_TEMUX_7_VT_GROUP; // setting for 7 x VT-Group
      otxRc = OtxTemuxSts1PayloadCfg(g_hTemux, nSlice, nSts1, eSts1Payload);
      DemoCheckDevRc(g_hTemux, otxRc);

      for (nVtGroup = 1; nVtGroup <= OTX_TEMUX_MAX_VTGROUP; nVtGroup++) {
          // configure each VT-Group for VT-1.5 payload
          eVtGroupPayload = OTX_TEMUX_VT_15; // setting for VT-1.5
          otxRc = OtxTemuxVtGroupPayloadCfg(g_hTemux, 
                                            nSlice, 
                                            nSts1, 
                                            nVtGroup, 
                                            eVtGroupPayload);
          DemoCheckDevRc(g_hTemux, otxRc);
      }
    }
  }

  //************************************************************************//    
  // PDH subsystem configuration
  //************************************************************************//

  // OTX_TEMUX_INTERFACE_HMVIP is supported
  // configure the system side interface for HMVIP
  // specify HMVIP mode of operation for system side interface.
  eSysIfMode = OTX_TEMUX_HMVIP;   
                  
  bHmvip_tfall = 0; // HMVIP_TFALL
  bHmvip_rfall = 0; // HMVIP_RFALL
  bUnused      = 0; // unused set to 0
                          
  otxRc = OtxTemuxSysIfModeCfg(g_hTemux, 
                               eSysIfMode, 
                               bUnused, 
                               bHmvip_tfall, 
                               bHmvip_rfall, 
                               bUnused, 
                               bUnused);  
  DemoCheckDevRc(g_hTemux, otxRc); 

  for (nSlice = 1; nSlice <= g_nLink; nSlice++){

    DemoLogf("OTX_TEMUX_T1_ESF %d\n", nSlice); 

    switch(g_demoEnv.eLinkSrcOrDest) {
    case OTX_TEMUX_BIT_ASYN:
      DemoLogf("OTX_TEMUX_BIT_ASYN %d\n", nSlice); 
      break;

    case OTX_TEMUX_BYTE_SYNC:
      DemoLogf("OTX_TEMUX_BYTE_SYNC %d\n", nSlice); 
      break;

    case OTX_TEMUX_TVT:
      DemoLogf("OTX_TEMUX_TVT %d\n", nSlice); 
      break;

    default:
      DemoLogf("   Not Configured OtxTemuxMapTribMode\n"); 
    }

    switch(g_demoEnv.eTxTimingMode) {
    case OTX_TEMUX_TX_TT_LOOP:
      DemoLogf("OTX_TEMUX_TX_TT_LOOP %d\n", nSlice); 
      break;

    case OTX_TEMUX_TX_TT_CTCLK:
      DemoLogf("OTX_TEMUX_TX_TT_CTCLK %d\n", nSlice); 
      break;

    case OTX_TEMUX_TX_TT_DSSI:
      DemoLogf("OTX_TEMUX_TX_TT_DSSI %d\n", nSlice); 
      break;

    case OTX_TEMUX_TX_TT_TISS:
      DemoLogf("OTX_TEMUX_TX_TT_TISS %d\n", nSlice); 
      break;

    case OTX_TEMUX_TX_TT_MUX:
      DemoLogf("OTX_TEMUX_TX_TT_MUX %d\n", nSlice); 
      break;

    default:
      DemoLogf("   Not Configured OtxTemuxTxTiming %d\n", nSlice); 
    }

    for(nSpe = 1; nSpe <= OTX_TEMUX_MAX_STS1; nSpe++){  
      // configure each slice and spe for E1 mode
      eSpeDest    = OTX_TEMUX_MAPPER; // T1 tribs associated with Mapper not MX13
      eSpePayload = OTX_TEMUX_T1;     // T1 tribs are specified not E1
      otxRc = OtxTemuxE1T1SpeCfg(g_hTemux, 
                                 nSlice, 
                                 nSpe, 
                                 eSysIfMode, 
                                 eSpeDest, 
                                 eSpePayload);
      DemoCheckDevRc(g_hTemux, otxRc); 
      // configure each slice and spe mapper to process data from the
      // T1E1 subsystem instead of the DS3E3 subsystem
      eSpeSource = OTX_TEMUX_T1E1;
      otxRc = OtxTemuxMapperSpeCfg(g_hTemux, nSlice, nSpe, eSpeSource);
      DemoCheckDevRc(g_hTemux, otxRc); 

      // T1 uses 28 tributaries
      for(nTrib = 1; nTrib <= OTX_TEMUX_MAX_TRIB; nTrib++){
        // configure each slice, spe and trib for T1 ESF
        eFrameMode = OTX_TEMUX_T1_ESF;  // T1 ESF
        // NOTE: UNFRAMED is not supported by the HMVIP interface
        otxRc = OtxTemuxE1T1TribFrameCfg(g_hTemux, 
                                         nSlice, 
                                         nSpe, 
                                         nTrib, 
                                         eFrameMode);
        DemoCheckDevRc(g_hTemux, otxRc); 
        // configure each slice, spe and trib ingress timing
        eBypass = OTX_TEMUX_ELST_DATAPATH; // ELST must be part of data path
        if(g_demoEnv.eLinkSrcOrDest == OTX_TEMUX_BIT_ASYN) {
          eRxTimingMode = OTX_TEMUX_RX_TT_RTDM; // bit asynchronously demapped
        }
        else {
          eRxTimingMode = OTX_TEMUX_RX_TT_EXSBI; // byte synchronously demapped
        }
        otxRc = OtxTemuxE1T1IngressTribTimingCfg(g_hTemux, 
                                                 nSlice, 
                                                 nSpe, 
                                                 nTrib, 
                                                 eBypass, 
                                                 eRxTimingMode);
        DemoCheckDevRc(g_hTemux, otxRc);
        // configure each slice, spe and trib egress timing
        eBypass = OTX_TEMUX_ELST_DATAPATH; // ELST must be part of data path
        // OTX_TEMUX_TX_TT_LOOP; OTX_TEMUX_TX_TT_CTCLK; CTCLK Internal Timing
        eTxTimingMode = g_demoEnv.eTxTimingMode;
        // trib destined for bit async mapping
        eLinkSrcOrDest = g_demoEnv.eLinkSrcOrDest;  
        otxRc = OtxTemuxE1T1EgressTribTimingCfg(g_hTemux, 
                                                nSlice, 
                                                nSpe, 
                                                nTrib, 
                                                eBypass, 
                                                eTxTimingMode,
                                                eLinkSrcOrDest);
        DemoCheckDevRc(g_hTemux, otxRc);
        // configure each slice, spe and trib for bit async or byte-sync mapping
        eLinkSrcOrDest = g_demoEnv.eLinkSrcOrDest; 
        ePdhTribType   = OTX_TEMUX_T1;       // T1 tributary
        eSonetTribType = OTX_TEMUX_VT_15;    // VT1.5 VT Type
        otxRc = OtxTemuxMapperTribCfg(g_hTemux, 
                                      nSlice, 
                                      nSpe, 
                                      nTrib, 
                                      eLinkSrcOrDest, 
                                      ePdhTribType, 
                                      eSonetTribType);
        DemoCheckDevRc(g_hTemux, otxRc);
      } // For(nTrib)
    } // For(nSpe)
  } // For(nSlice)
}


/*
**
** Burst setup and processing functions
**
*/

/*******************************************************************************
  @func
    Process burst receive event for one highway
*******************************************************************************/
void *DemoBurstReceiveProc(
  void *lpParameter   // thread data
)
{
  OTX_RESULT otxRc;

  OtxEventDataS eventData;
  OTX_INT32 nHighway;
  OTX_INT32 nBufferCnt;
  OTX_INT32 nFifoSize;
  OTX_INT32 nMaxFrameCnt;

  nHighway = (OTX_INT32) *((OTX_INT32 *) lpParameter);

  DemoLogf("DemoBurstReceiveProc: Highway %d\n", nHighway);

  otxRc = OtxDrvGetAttributeValueUint32(g_hBurst, 
                                        OTX_ATTR_BURST_RECEIVER_FIFO_SIZE,           
                                        &nFifoSize); 
  DemoCheckDevRc(g_hBurst, otxRc);

  nMaxFrameCnt = nFifoSize/DEMO_BURST_BUF_SIZE;

  while (g_bDemoRunnig) {
    OtxDrvWaitForSingleEvent(g_hBurstReceiveEvent[nHighway], 
                             OTX_TIME_INFINITY);

    do {
      otxRc = OtxDrvGetEventData(g_hBurstReceiveEventQueue[nHighway], 
                                  &eventData); 
      DemoCheckRc(otxRc);
      if(OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        // We have a event make sure it is for the burst receive
        if((eventData.m_eDeviceType == OTX_DEVICE_BURST) &&
           (eventData.m_nCode == OTX_BURST_RECEIVER_EC_DATA_AVAILABLE)) {
          // Process the event for one highway only check to make sure 
          // the event is for the this highway.
          if(eventData.m_nRequestId == nHighway) {
            OTX_INT32 nFullBuffers;;
            // number of bytes of data in host memory
            nBufferCnt = eventData.m_nParam;
            if(nBufferCnt < DEMO_BURST_BUF_SIZE) {
              DemoLogf("DemoBurstReceiveProc(%d) nBufferCnt %d should be at least %d\n", 
                        nHighway, 
                        nBufferCnt,
                        DEMO_BURST_BUF_SIZE);
            }

            nFullBuffers = nBufferCnt/DEMO_BURST_BUF_SIZE;
            if(nFullBuffers > nMaxFrameCnt/2) {
              // There is a lot of data in host buffer with 50% or more of our 
              // buffer available for transfer. Process multiple this time.
              if(g_hLogicDevs.mBurstReceiver[nHighway].bActive) {
                nFullBuffers = nFullBuffers/2;
              }
              else {
                // not a highway that is currently being used
                nFullBuffers = 0;
              }
            }
            else {
              if(g_hLogicDevs.mBurstReceiver[nHighway].bActive) {
                // Just get one buffer
                nFullBuffers = 1;
              }
              else {
                // not a highway that is currently being used
                nFullBuffers = 0;
              }
            }
            // Get all the data out of host memory
            while(nFullBuffers) {
              DemoBurstReceiver(eventData.m_nRequestId, // Highway
                                &nBufferCnt);           // buffer byte count
              if(nBufferCnt < DEMO_BURST_BUF_SIZE) {
                // Not enough for a full buffer of data
                nFullBuffers = 0;
              }
              else {
                nFullBuffers--;
              }
            }
          }
          else {
            // It is for another highway
            DemoLogf("DemoBurstReceiveProc(%d) Bad highway number %d\n", 
                      nHighway,        
                      eventData.m_nRequestId);
          }
        }
        else {
          DemoLogf("DemoBurstReceiveProc[%d]: Unknown event type %d, code %d\n",
                   nHighway,
                   eventData.m_eDeviceType,
                   eventData.m_nCode);
        }
      }
    } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
  }

  if(OTX_RESULT_CODE(otxRc) != OTX_S_OK) {
    DemoLogf("DemoBurstReceiveProc[%d]:: Event OTX error 0x%X\n", 
             nHighway, 
             otxRc);
  }
  else {
    DemoLogf("DemoBurstReceiveProc[%d]:: thread done\n", nHighway);
  }

  return 0;
}


/*******************************************************************************
  @func
    Burst device test function using a fixed pattern
*******************************************************************************/
void DemoBurstTestFixedData(OTX_UINT32 nHwy)
{
  OTX_RESULT       otxRc;
  OTX_RESULT       otxRcStart;
  OTX_RESULT       otxRcFinish;
  OTX_UINT32       nHwyLast;
  OtxDeviceStateE  eDevState;
  OTX_DATETIME     nTimeStampStart;
  OTX_DATETIME     nTimeStampFinish;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckDevRc(g_hBurst, otxRc);

  if((OTX_INT32) nHwy >= g_nActiveHighways) {
    DemoLogf("Burst highway %d not active\n", nHwy);
  }

  otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
  DemoCheckDevRc(g_hTemux, otxRc);

  nHwyLast = g_nActiveHighways - 1;

  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Burst device is already enable\n");
  }
  else {
    // wait to last highway to enable the burst device
    if(nHwy == nHwyLast) {
      // By enabling the Burst device, we start the process of transferring 
      // data from the hardware to the driver via DMA.  The data is stored 
      // in the Burst device's buffer for processing later.
      DemoLogf("Enable g_hBurst\n");
      otxRcStart = OtxDrvGetCurrentTime(&nTimeStampStart); 
      otxRc = OtxDrvEnable(g_hBurst, OTX_TASK_SYNC, 5000L);
      otxRcFinish = OtxDrvGetCurrentTime(&nTimeStampFinish); 
      DemoCheckDevRc(g_hBurst, otxRc);
      // Check these here so any time needed to perform the test is not
      // part of the check.
      DemoCheckRc(otxRcStart);
      DemoCheckRc(otxRcFinish);
      DemoLogf("Enable nTimeStampStart\n");
      DemoPrintTimeStamp(nTimeStampStart);
      DemoLogf("\n");
      DemoLogf("Enable nTimeStampFinish\n");
      DemoPrintTimeStamp(nTimeStampFinish);
      DemoLogf("\n");
    }
  }

  if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_FALSE) {
    DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", 
              nHwy);
    g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter = 0;
    g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_TRUE;
  }
  else {
    DemoLogf("Burst Receiving Hwy[%d] already running.\n", nHwy);
  }

  if (g_hLogicDevs.mBurstSender[nHwy].bActive == OTX_FALSE) {
    DemoLogf("Burst Sender Hwy[%d]. Sending a fixed pattern continuously\n", 
              nHwy);

    g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;

    g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = 0;

    g_hLogicDevs.mBurstSender[nHwy].bActive = OTX_TRUE;

    g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = DEMO_BURST_BUF_SIZE;
  } 
  else {
    DemoLogf("Burst Sender Hwy[%d] already running.\n", nHwy);
  }

  if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
    // Burst receiver is running. Ok.

    if (!g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode &&
        !g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
      // Fixed mode is not enabled. So enable it.
  
      g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode = OTX_TRUE;
      DemoLogf("Burst Receiver Hwy[%d] Fixed Test Mode is now enabled\n", 
               nHwy);

    } else {
      DemoLogf("Burst Receiver Hwy[%d] Test Mode already running\n", nHwy);
    }
  } 
  else {
    DemoLogf("Burst Receiver Hwy[%d] The Burst Receiver is not running\n", 
             nHwy);
  }

  // wait to last highway to enable the burst device
  if(nHwy == nHwyLast) {
    // We just starrted to send valid GAMMA data, but it will take
    // a little time to fill all the output and then input buffer
    // with the data.
    OsaThreadSleep(100);
  }
}


/*******************************************************************************
  @func
    Burst device test function using GAMMA pattern (1 Frame)
*******************************************************************************/
void DemoBurstTestGammaData(OTX_UINT32 nHwy)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  i;
  OTX_UINT32  nHwyLast;
  OtxDeviceStateE  eDevState;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckDevRc(g_hBurst, otxRc);

  if((OTX_INT32)nHwy >= g_nActiveHighways) {
    DemoLogf("Burst highway %d not active\n", nHwy);
  }

  otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
  DemoCheckDevRc(g_hTemux, otxRc);

  nHwyLast = g_nActiveHighways - 1;

  if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_FALSE) {
    DemoLogf("Burst Receiver Hwy[%d]: Receiving data (no file logging).\n", nHwy);
    g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter = 0;
    g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_TRUE;
  }
  else {
    DemoLogf("Burst Receiving Hwy[%d] already running.\n", nHwy);
  }

  if (g_hLogicDevs.mBurstSender[nHwy].bActive == OTX_FALSE) {
    DemoLogf("Burst Sender Hwy[%d]. Sending a gamma pattern continuously\n", nHwy);

    g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;

    // The starting point for the gamma pattern will be value=highway
    g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = 0;
    for (i = 0; i < (OTX_UINT32) DEMO_BURST_TS_SIZE; i++) {
      g_hLogicDevs.mBurstSender[nHwy].sqGammaData[i] = (unsigned char) nHwy;
    }

    g_hLogicDevs.mBurstSender[nHwy].bActive = OTX_TRUE;

    g_hLogicDevs.mBurstSender[nHwy].cbTxBuf = DEMO_BURST_BUF_SIZE;
  } 
  else {
    DemoLogf("Burst Sender Hwy[%d] already running.\n", nHwy);
  }

  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Burst device is already enable\n");
  }
  else {
    // wait to last highway to enable the burst device
    if(nHwy == nHwyLast) {
      // By enabling the Burst device, we start the process of transferring 
      // data from the hardware to the driver via DMA.  The data is stored 
      // in the Burst device's buffer for processing later.
      DemoLogf("Enable g_hBurst\n");
      otxRc = OtxDrvEnable(g_hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckDevRc(g_hBurst, otxRc);
    }
  }

  if (g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
    // Burst receiver is running. Ok.

    if (!g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode &&
        !g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode) {
      // Gamma mode is not enabled. So enable it.
      int n;

      for (n = 0; n < DEMO_BURST_TS_SIZE; n++) {
        g_hLogicDevs.mBurstReceiver[nHwy].sqGammaTestData[n]  = 0;
        g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[n] = 0;
      }
  
      g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_TRUE;
      DemoLogf("Burst Receiver Hwy[%d] Gamma Test Mode is now enabled\n", 
                nHwy);

    } else {
      DemoLogf("Burst Receiver Hwy[%d] Test Mode already running\n", 
                nHwy);
    }
  } 
  else {
    DemoLogf("Burst Receiver Hwy The Burst Receiver is not running\n", 
              nHwy);
  }

  // wait to last highway to enable the burst device
  if(nHwy == nHwyLast) {
    // We just starrted to send valid GAMMA data, but it will take
    // a little time to fill all the output and then input buffer
    // with the data.
    OsaThreadSleep(100);
  }
}


/*******************************************************************************
  @func
    Burst Device Test Stop all Testing
*******************************************************************************/
void DemoBurstEndTest(void)
{
  OTX_RESULT  otxRc;
  OtxDeviceStateE  eDevState;
  OTX_UINT32 nHwy;

  otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
  DemoCheckDevRc(g_hBurst, otxRc);

  if (eDevState == OTX_DEVICE_WORKING) {
    DemoLogf("Disable g_hBurst\n");
    otxRc = OtxDrvDisable(g_hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckDevRc(g_hBurst, otxRc);
  }
  else {
    DemoLogf("Burst device is already disabled\n");
  }

  for(nHwy = 0; nHwy < DEMO_MAX_HWY_NO; nHwy++) {
    g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode = OTX_FALSE;
    g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode = OTX_FALSE;

    if (g_hLogicDevs.mBurstReceiver[nHwy].bActive) {
      if (g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter == 0) {
        DemoLogf("Error: No burst data was received, Highway[%d]\n", nHwy);
      }
    }

    g_hLogicDevs.mBurstReceiver[nHwy].bActive = OTX_FALSE;
    if (g_hLogicDevs.mBurstReceiver[nHwy].pRxFile) {
      DemoLogf("BurstReceiver[%d]: Closing file\n", nHwy);
      fclose(g_hLogicDevs.mBurstReceiver[nHwy].pRxFile);
      g_hLogicDevs.mBurstReceiver[nHwy].pRxFile = NULL;
    }

    g_hLogicDevs.mBurstSender[nHwy].bActive   = OTX_FALSE;
    if (g_hLogicDevs.mBurstSender[nHwy].pTxFile) {
      DemoLogf("BurstSender[%d]: Closing file\n", nHwy);
      fclose(g_hLogicDevs.mBurstSender[nHwy].pTxFile);
      g_hLogicDevs.mBurstSender[nHwy].pTxFile = NULL;
    }
  }
}


/*******************************************************************************
  @func
    Check frame refence ID for this buffer of data
*******************************************************************************/
void DemoBurstCheckRefenceId(OTX_UINT64 nFrameRefId, OTX_UINT32 nSize, OTX_UINT32 nHwy)
{
  OTX_UINT32 nTestValue;
  OTX_UINT64 nRange;
  OTX_UINT64 nMaskValue = 0xFFFFFFFF;
  OTX_UINT64 nChkFrameRefId;

  nChkFrameRefId = g_hLogicDevs.mBurstReceiver[nHwy].nFrameRefId;
  // Check the frame refence ID. It should be different for each frame.
  if(nChkFrameRefId >= nFrameRefId) {
    // frame reference should be incrementing
    DemoLogf("Burst[%d] nFrameRefId did not increment previous value\n", 
             nHwy);

    DemoLogf("  Previous frame ID %llu\n", 
             nChkFrameRefId);
    nTestValue = (OTX_UINT32) (nChkFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nChkFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", nTestValue);


    DemoLogf("  This frame ID %llu\n", nFrameRefId);
    nTestValue = (OTX_UINT32) (nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", nTestValue);


    if(nFrameRefId == 0) {
      // Counter is set to zero rolled over
      DemoLogf("Burst[%d] ID is zero\n", nHwy);
    }
    if(nFrameRefId == nChkFrameRefId) {
      // The ID is the same
      DemoLogf("Burst[%d] Frame Reference ID is equal\n", nHwy);
    }
  }

  nRange = (nSize/DEMO_BURST_TS_SIZE);
  if(nChkFrameRefId > (nFrameRefId - nRange)) {
    // The buffer has multiple frames
    DemoLogf("Burst[%d] nFrameRefId range is not correct for buffer size %d\n",
             nHwy,
             nSize);
    DemoLogf("  Previous frame ID %llu, this frame ID %llu\n", 
             nChkFrameRefId,
             nFrameRefId);

    DemoLogf("  Previous frame ID %llu\n", 
             nChkFrameRefId);
    nTestValue = (OTX_UINT32) (nChkFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nChkFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);

    DemoLogf("  This frame ID %llu\n", 
             nFrameRefId);
    nTestValue = (OTX_UINT32) (nFrameRefId & nMaskValue);
    DemoLogf("Lower bits of nFrameRefId 0x%X, mask %llu\n", 
             nTestValue,
             nMaskValue);
    nTestValue =  (OTX_UINT32) (nFrameRefId >> 32);
    DemoLogf("Upper bits of nFrameRefId 0x%X\n", 
             nTestValue);
  }
  g_hLogicDevs.mBurstReceiver[nHwy].nFrameRefId = nFrameRefId;
}


/*******************************************************************************
  @func
    Check Fixed data pattern for errors; pattern repeats for each frame.
*******************************************************************************/
void DemoBurstCheckFixedPattern(OTX_UINT32 *pInputData, OTX_UINT32 nHwy)
{
  OTX_UINT32 nY;
  char       nCheckValue;
  char       *pData;

  pData = (char *) pInputData;
  if(g_nPattern == 8) {
    // This pattern generated by the FPGA.
    // Check counter, every frame has a different starting count
    // But every frame should always start with the same value
    // Use the first channel as the starting point.
    nCheckValue = pData[0];
  }
  else
  if(g_nPattern == 0xA) {
    // This is an STM-1 data pattern generated by test equipment.
    // Every frame has a different data pattern. But every frame will 
    // use the same value for all timeslots. TS0 is framing data.
    // Pick a channel other then zero to test all other channels. 
    nCheckValue = pData[2];
  }
  else {
    // used pattern looded into send buffer
    nCheckValue = (char) g_sqSendFixedBuf[nHwy][0];
  }

  // Check one frame of data at a time
  for(nY=0; nY < DEMO_BURST_TS_SIZE; nY++) {
    if((*pData != nCheckValue) && (nY!=0)) {
      if((g_hLogicDevs.mBurstReceiver[nHwy].bDisplayDataErr)) {
        // Display the first error
        DemoLogf("Receive frame Hwy[%d] cnt[%d]: Expected 0x%X was set to 0x%X\n", 
                 nHwy, 
                 nY, 
                 ((OTX_UINT32)nCheckValue & 0xFF), 
                 ((OTX_UINT32)*pData & 0xFF));

        DemoLogf("g_FetchCount %d, g_FetchCountCur %d, nDataBlockCounter %d\n",  
                 g_FetchCount,
                 g_FetchCountCur,
                 g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter);
        
        DemoLogf("nY %d, nHwy %d, g_sqSendFixedBuf[nHwy][nY] 0x%X\n",  
                 nY,
                 nHwy,
                 g_sqSendFixedBuf[nHwy][nY]);
        
        DemoLogf("\n");

        // Only display the first error. Printing a lot to the display
        // can cause the burst buffers to overflow
        g_hLogicDevs.mBurstReceiver[nHwy].bDisplayDataErr = OTX_FALSE;
      }
      g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[nY]++;
    }
    pData++;
    if(g_nPattern == 8) {
      nCheckValue++;
    }
    else {
      // Pattern 0xA does not change
      if(g_nPattern != 0xA) {
        nCheckValue = (char) g_sqSendFixedBuf[nHwy][nY + 1];
      }
    }
  }
}


/*******************************************************************************
  @func
    Check Gamma data pattern for errors (one frame)
*******************************************************************************/
void DemoCheckBurstGammaData(OTX_UINT8 *pBuf, OTX_UINT32 *pErrCnt, OTX_UINT32 nHighway)
{
  OTX_UINT32 nTimeSlot;
  
  OTX_UINT8  *pData;
  OTX_UINT8  *pBufData;

  OTX_UINT32 nDisplayCnt = 0;

  pBufData = pBuf;

  // GAMMA pattern: update with each frame
  pData = &g_hLogicDevs.mBurstReceiver[nHighway].sqGammaTestData[0];

  for(nTimeSlot = 0; nTimeSlot < DEMO_BURST_TS_SIZE; nTimeSlot++) {
    if(g_hLogicDevs.mBurstReceiver[nHighway].bTestReset == OTX_TRUE) {
      // User wants to reset the error count and start value
      *pData = *pBufData;
      g_hLogicDevs.mBurstReceiver[nHighway].nGammaTestErrors[nTimeSlot] = 0;
    }
    else
    if(*pBufData != *pData) {
      // Data mismatch. Step error count and resync
      if(g_hLogicDevs.mBurstReceiver[nHighway].bDisplayDataErr) {

        // Found a buffer with bad data
        DemoLogf("Err: Hwy(%d) Expected 0x%X Received 0x%X, Block Counter %d\n",
                 nHighway,
                 *pData,
                 *pBufData,
                 g_hLogicDevs.mBurstReceiver[nHighway].nDataBlockCounter);
        DemoLogf("Time slot %d \n", nTimeSlot);
        DemoLogf("g_FetchCount %d, g_FetchCountCur %d,  nGammaZeroXferCnt %d\n",
                 g_FetchCount,
                 g_FetchCountCur,
                 g_hLogicDevs.mBurstSender[nHighway].nGammaZeroXferCnt);


        g_hLogicDevs.mBurstReceiver[nHighway].bDisplayDataErr = OTX_FALSE;
      }
      *pErrCnt = *pErrCnt + 1; // total number of errors
      // Data was in error, next check will be based on this value
      *pData = *pBufData;
      g_hLogicDevs.mBurstReceiver[nHighway].nGammaTestErrors[nTimeSlot]++;
    }
    // Step the GammaTestData value, which is the next value that we expect
    *pData += DEMO_GAMMA_INCREMENT;
  
    pData++;
    pBufData++;
  } // For time slot
}


/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId, OTX_UINT32 nBlockSize)
{

#if 1
        
  DemoLogf("DemoBurstSender: not supported at this time nSourceId(%d)!\n", 
           nSourceId);
    
  return;

#else
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;
  OTX_UINT32  nFrameSize;

  nFrameSize = DEMO_BURST_TS_SIZE;

  // validate parameters
  if ((nSourceId < 0) || ((OTX_INT32) nSourceId >= g_nActiveHighways)) {
    if ((nSourceId < 0) || (nSourceId >= DEMO_T1_MAX_HWY_NO)) {
      DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    }
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  if(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf != DEMO_BURST_BUF_SIZE) {
    // Demo setup issue the buffer should hold the data.
    DemoLogf("DemoBurstSender[%d]: Transfer buffer is not correct size %d\n",
              nSourceId,
              g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);

    if(g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
      DemoLogf("DemoBurstSender[%d]: Closing file\n", nSourceId);
      fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
      g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
    }
    DemoLogf("DemoBurstSender[%d]: Burst device is no longer active\n",
             nSourceId);
    g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;

    return;
  }

  if(g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    int nItemSize;
    int cbTsBuf;
    int cbActualTsBuf;
        
    OTX_UINT32 nDataPad;
    OTX_UINT32 nByteCnt;
    OTX_UINT32 nTxBufSize;
    OTX_UINT32 nTxDataSize;


    // This is bigger than we actually use
    OTX_UINT8  singleTsBuf[DEMO_BURST_BUF_SIZE]; 
    OTX_UINT8  *pBuffer;

    nTxBufSize = sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf);
    nTxDataSize = sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]);
    nByteCnt = nTxBufSize / nTxDataSize;

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS are being sent
      pBuffer = g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf;
      nItemSize =  sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]);
      cbTsBuf = nByteCnt;
    }
    else {
      // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
      memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, nByteCnt);
      // Single TS
      pBuffer = singleTsBuf;
      nItemSize =  sizeof(singleTsBuf[0]);
      cbTsBuf = DEMO_BURST_BUF_SIZE / nFrameSize;
    }

    // Get data from the file
    cbActualTsBuf = fread(
      pBuffer,
      nItemSize,
      cbTsBuf,
      g_hLogicDevs.mBurstSender[nSourceId].pTxFile
      );

    if(cbActualTsBuf == 0) {
      // End of file no data was read from file
      DemoLogf("DemoBurstSender[%d]: File transmission complete\n", nSourceId);
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        // Start file from begining
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Starting file from the beginning.\n", 
                   nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);

        // Now get the data from file
        cbActualTsBuf = fread(
          pBuffer,
          nItemSize,
          cbTsBuf,
          g_hLogicDevs.mBurstSender[nSourceId].pTxFile
          );
            
        if(cbActualTsBuf == 0) {
          // We seem to still have no data
          DemoLogf("DemoBurstSender[%d]: no data found in file\n");

          if(g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
            DemoLogf("DemoBurstSender[%d]: Closing file\n", nSourceId);
            fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
            g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
          }
          DemoLogf("DemoBurstSender[%d]: Burst device is no longer active\n");
          g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;

          return;
        }

      } 
      else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        DemoLogf("DemoBurstSender[%d]: Closing file\n", nSourceId);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
        // Our work here is done no more data to send
        return;
      }
    }
        
    nDataPad = (cbTsBuf - cbActualTsBuf) / nItemSize;

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
              
      if ((cbActualTsBuf > 0) && (nDataPad)) {
        // We are at the end of the file, and we don't have a full 
        // buffer to send. Pad with 0xFF at the end of the buffer
        DemoLogf("Pad with 0xFF [%d] at the end of the buffer\n", nDataPad);
      
        pBuffer = &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0];

        memset(&pBuffer[cbActualTsBuf], 0xFF, nDataPad);

        cbActualTsBuf = nByteCnt;
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbActualTsBuf;
    }
    else {
      // Single TS
      int cbSingleTsBuf;
      int x;
      int cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / nFrameSize;
      int byteCount = 0;
      
      cbSingleTsBuf = cbActualTsBuf;

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (nDataPad)) {
          // We are at the end of the file, and we don't have a full 
          // buffer to send. Pad with 0xFF at the end of the buffer 
          // to make sure that we don't underrun the Burst Sender
          DemoLogf("Pad with 0xFF [%d] at the end of the buffer\n", nDataPad);
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, nDataPad);
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend;
        while((byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE)) {
          g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount];
          byteCount++;
          x += nFrameSize;
        }
        
        //assert(byteCount == cbSingleTsBuf);
        if(byteCount == cbSingleTsBuf) {
          // Demo setup issue the buffer should hold the data.
          DemoLogf("DemoBurstSender[%d]: Single TS buffer byte count in error\n");

          if(g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
            DemoLogf("DemoBurstSender[%d]: Closing file\n", nSourceId);
            fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
            g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
          }
          DemoLogf("DemoBurstSender[%d]: Burst device is no longer active\n");
          g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;

          return;
        }
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * nFrameSize;
    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      otxRc = OtxBurstSendData(g_hBurst, 
                               nSourceId, 
                               g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, 
                               &ui32);
      DemoCheckDevRc(g_hBurst, otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: sent a short buffer %d, expected %d\n", 
          nSourceId, 
          ui32, 
          g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } 
    else {
      // Demo setup issue the buffer should have data.
      DemoLogf("DemoBurstSender[%d]: No data in burst output buffer\n");

      if(g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
        DemoLogf("DemoBurstSender[%d]: Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
      }
      DemoLogf("DemoBurstSender[%d]: Burst device is no longer active\n");
      g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
    }
    
  } 
  else {
    // Send a gamma or fixed pattern. GAMMA means that we increase the 
    // value by one for each frame (continuously until aborted)
    OTX_UINT32 x;
    OTX_UINT32 y;
    OTX_UINT8  *pDataTo;
    OTX_UINT8  *pDataFrom;
    OTX_UINT8  sqDataStart[DEMO_BURST_BUF_SIZE];
    OTX_BOOL   bGamma = OTX_FALSE;

    pDataTo = &sqDataStart[0];
    pDataFrom = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
    for (y = 0; y < nFrameSize; y++) {
      *pDataTo = *pDataFrom;
      pDataTo++;
      pDataFrom++;
    }

    pDataTo = &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0];

    if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {
      pDataFrom = g_sqSendFixedBuf[nSourceId];
    }
    else {
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
        bGamma = OTX_TRUE;
      }
      else {
        // Need to send something send fixed
        pDataFrom = g_sqSendFixedBuf[nSourceId];
      }
    }

    for (x = 0; x < DEMO_BURST_BUF_SIZE; x += nFrameSize) {
      // Set the frame data pointer, all channel are updated
      if(bGamma) {
        pDataFrom = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
      }
      for (y = 0; y < nFrameSize; y++) {
        *pDataTo = *pDataFrom;
        if (bGamma) {
          *pDataFrom = *pDataFrom + DEMO_GAMMA_INCREMENT;
        }
        pDataTo++;
        pDataFrom++;
      }
    }

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hBurst, 
                             nSourceId, 
                             g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, 
                             &ui32);
    if(OTX_FAILURE(otxRc)) {
      // Add to logfile do not print to console. The console display takes
      // too much CPU time resulting in the DMA buffer overflows
      DemoLogfile("DemoBurstSend ERROR(%#lX [%d.%d])\n", 
                  OTX_RESULT_CODE(otxRc), 
                  OTX_RESULT_SOURCE_ID(otxRc), 
                  OTX_RESULT_LINE_NUM(otxRc));
    }

    // check the amount buffered w/ how much we requested to be buffered
    // (since we set the trigger amount to the buffer size, there should
    // have been room to buffer this, so these two should be the same).
    //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
    if(ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
      if((ui32 == 0) && (bGamma)) {
        // sent no data
        OTX_UINT32 y;
        pDataTo = &g_hLogicDevs.mBurstSender[nSourceId].sqGammaData[0];
        pDataFrom = &sqDataStart[0];
        for (y = 0; y < nFrameSize; y++) {
          *pDataTo = *pDataFrom;
          pDataTo++;
          pDataFrom++;
        }
        g_hLogicDevs.mBurstSender[nSourceId].nGammaZeroXferCnt++;
      }
      else {
        DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer.\n", nSourceId);
        DemoLogf("   Buffer size sent %d\n", g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
        DemoLogf("   Buffer size returned %d\n", ui32);
        DemoLogf("g_FetchCount %d, g_FetchCountCur %d, nBlockSize %d\n",  
                 g_FetchCount,
                 g_FetchCountCur,
                 nBlockSize); 
      }
    }
  }

  return;
#endif
}


/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId, OTX_UINT32 *pBlockSize)
{
  OTX_UINT32    ui32;
  OTX_UINT32    nBufferSize;
  OTX_UINT32    nFrameSize;

  OTX_UINT8     *pBuf;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  OTX_UINT32    nLoopErrCnt = 0;
  OTX_UINT32    nLoopDisplayErrCnt = 0;
  OTX_UINT32    nDisplayCnt = 0;
  OTX_BOOL      bGamma = OTX_FALSE;
  OTX_BOOL      bDisplayBuf = OTX_FALSE;
  OTX_RESULT    otxRc = OTX_S_OK;

  static OTX_UINT8 HwyBuf[DEMO_T1_MAX_HWY_NO][DEMO_BURST_BUF_SIZE];

  nFrameSize = DEMO_BURST_TS_SIZE;

  // validate parameters
  if((nSourceId < 0) || ((OTX_INT32) nSourceId >= g_nActiveHighways)) {
    // if it is not an active highway just return
    if ((nSourceId < 0) || (nSourceId >= DEMO_T1_MAX_HWY_NO)) {
      DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    }
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter++;
  // Each highway has a buffer get this highway's buffer address
  pBuf = &HwyBuf[nSourceId][0];

  memset(pBuf, 0, DEMO_BURST_BUF_SIZE);

  // we want the frame reference id and time
  ui32 = DEMO_BURST_BUF_SIZE;
  otxRc = OtxBurstReadDataEx(g_hBurst, 
                             nSourceId, 
                             pBuf, 
                             &ui32, 
                             &nFrameRefId, 
                             &dtFrameRefTime);
  if(OTX_FAILURE(otxRc)) {
    if(OTX_RESULT_CODE(otxRc) == OTX_S_FALSE) {
      // This could be an over/under run issue
      // The console display takes too much CPU time and 
      // then the DMA buffer overflow is increased.
      g_hLogicDevs.mBurstReceiver[nSourceId].nOverflowErrCnt++;
    }
    else {
      // This is not an not an over/under run issue
      DemoLogf("DemoBurstReceiver ERROR(%#lX [%d.%d])\n", 
                OTX_RESULT_CODE(otxRc), 
                OTX_RESULT_SOURCE_ID(otxRc), 
                OTX_RESULT_LINE_NUM(otxRc));
    }
  }

  // check the amount buffered w/ how much we requested to be buffered
  // (since we set the trigger amount to the buffer size, there should
  // have been room to buffer this, so these two should be the same).
  if(ui32 != DEMO_BURST_BUF_SIZE) {
    *pBlockSize = ui32;
    if(ui32 == 0) {
      // Buffer empty no data to process
      return;
    }
    // Process what we received

    /*
    DemoLogf("Burst[%d] receive a short buffer size %d should be %d\n", 
             nSourceId, 
             ui32, 
             DEMO_BURST_BUF_SIZE);
    */
  }

  if(ui32 > DEMO_BURST_BUF_SIZE) {
    DemoLogf("Burst[%d] receive size %d is larger then buffer %d\n", 
             nSourceId, 
             ui32, 
             DEMO_BURST_BUF_SIZE);

    ui32 = DEMO_BURST_BUF_SIZE;
  }

  // We do not always get a full buffer
  nBufferSize = ui32;
  // Every frame in the buffer will need data, 
  // check for a partial frame.
  if(nBufferSize % nFrameSize) {
    DemoLogf("DemoBurstReceiver[%d] Buffer contains a partial frame\n", 
             nSourceId);
    DemoLogf("DemoBurstReceiver[%d] Buf size %d, Frame size %d\n", 
             nSourceId, 
             nBufferSize,
             nFrameSize);
    DemoLogf("g_FetchCount %d, g_FetchCountCur %d, nBlockSize %d\n",  
             g_FetchCount,
             g_FetchCountCur,
             *pBlockSize); 

    // The size needs to be a multiple of the timeslots
    nBufferSize = (nBufferSize/nFrameSize) * nFrameSize;
  }

  // Each frame should have a unique reference ID
  DemoBurstCheckRefenceId(nFrameRefId, nBufferSize, nSourceId);

  // Check the received value if we are in Gamma or fixed test mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode ||
      g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
    OTX_UINT32 nFrame;
    OTX_BOOL   bDisplayDataErr;
    OTX_UINT8  *pBufData;

    nLoopErrCnt = 0;
    bDisplayDataErr = g_hLogicDevs.mBurstReceiver[nSourceId].bDisplayDataErr;

    pBufData    = pBuf;
    if (g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {     
      // Fixed data pattern is provided by hardware and is a
      // byte count sequence.
      bGamma = OTX_FALSE;
    }
    else {
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bGammaTestMode) {
        bGamma = OTX_TRUE;
      }
    }
    
    for (nFrame = 0; nFrame < nBufferSize; nFrame += nFrameSize) {
      if(g_hLogicDevs.mBurstReceiver[nSourceId].bFixedTestMode) {
        if(g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset == OTX_FALSE) {
          OTX_UINT32 *pFrame;
          pFrame = (OTX_UINT32 *) (pBufData + nFrame);
          DemoBurstCheckFixedPattern(pFrame, nSourceId);
        }
        else {
          OTX_UINT32 nTimeSlot;
          for (nTimeSlot = 0; nTimeSlot < DEMO_BURST_TS_SIZE; nTimeSlot++) {
            // User wants to reset the error count 
            g_hLogicDevs.mBurstReceiver[nSourceId].nGammaTestErrors[nTimeSlot] = 0;
          }
        }
      }
      else {
        DemoCheckBurstGammaData((pBufData + nFrame), &nLoopErrCnt, nSourceId);
        if((nLoopErrCnt) && 
           (bDisplayDataErr) && 
           (nLoopDisplayErrCnt == 0)) {
          // Display this buffer error location and data before the error
          if(nFrame != 0) {
            nDisplayCnt = nFrame - nFrameSize;
            // save this value so it can be displayed.
            nLoopDisplayErrCnt = nLoopErrCnt;

            bDisplayBuf = OTX_TRUE;
          }
          else {
            // Error at the start of buffer could be result of last buffer
            // that was checked. Do not log this error look for one within
            // this buffer of data.
            g_hLogicDevs.mBurstReceiver[nSourceId].bDisplayDataErr = OTX_TRUE;
          }

          DemoLogf("\nReceiver error %d, frame %d, offset %d, buffer size %d\n", 
                   nLoopErrCnt, 
                   nFrame, 
                   nDisplayCnt,
                   nBufferSize);
        }
      }
    } // For each frame
  }

  if(g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset == OTX_TRUE) {
    g_hLogicDevs.mBurstReceiver[nSourceId].bTestReset = OTX_FALSE;
    g_hLogicDevs.mBurstReceiver[nSourceId].nDataBlockCounter = 0;
    g_hLogicDevs.mBurstReceiver[nSourceId].nOverflowErrCnt = 0;

    DemoLogf("Burst Hwy[%d] Reset Fixed data and GAMMA test\n", nSourceId);
  }

  // Echo if we're in verbose mode or
  // if the data check has found an error display it
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose || bDisplayBuf) {
    OTX_UINT32 x;
    OTX_UINT32 nEndCnt;
    // Display one buffer only, set above if error is detected
    bDisplayBuf = OTX_FALSE;
    if(nLoopDisplayErrCnt) {
      DemoLogf("Burst Receiver frame error count %d, buffer error count %d\n", 
                nLoopDisplayErrCnt,
                nLoopErrCnt);
    }
    DemoPrintTimeStamp(dtFrameRefTime);

    // Only display the least significant 32 bits of the frame reference id
    DemoLogf("Burst Receiver Hwy[%d] (len=%.5d) ID:%08x, Starting Offset 0x%X\n", 
              nSourceId, 
              nBufferSize, 
              (OTX_UINT32) (nFrameRefId & 0xffffffff),
              nDisplayCnt);

    // Normally it starts at frame 0, but an error above can set it
    // to the frame that was in error. 
    // Print one frame of data.
    if(nDisplayCnt % nFrameSize) {
      DemoLogf("DemoBurstReceiver[%d] display count not on frame boundary %d\n", 
               nSourceId, nDisplayCnt);
      nDisplayCnt = (nDisplayCnt/nFrameSize) * nFrameSize;
    }
    x = nDisplayCnt;
    nEndCnt = x + (nFrameSize * 50);
    while( (x < nEndCnt) && 
           (x < DEMO_BURST_BUF_SIZE) && 
           (x < nBufferSize) ) {
      DemoLogf("%.2X ", pBuf[x]);
      x++;
      if(!(x%16)) {
        DemoLogf("\n"); 
      }
    }

    DemoLogf("\n");
  }
  
  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile && 
      g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(pBuf, 
                         nBufferSize, 
                         nBufferSize, 
                         g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if ((cbWritten != (int) DEMO_BURST_BUF_SIZE) && (!g_nBurstExitFlag)) {
        DemoLogf("Error:Receiver[%d] Could not write frame to file\n", nSourceId);
        DemoLogf("  size=%d, cbWritten=%d\n", DEMO_BURST_BUF_SIZE, cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      OTX_UINT32 x;

      byteCount = 0;
      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < nBufferSize; 
          x += nFrameSize
      ) {
        singleTsBuf[byteCount++] = pBuf[x];
      }
      cbWritten = fwrite(singleTsBuf, 
                         sizeof(singleTsBuf[0]), 
                         byteCount, 
                         g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if ((cbWritten != byteCount) && (!g_nBurstExitFlag)) {
        DemoLogf("Error:Burst Receiver[%d] Could not write data to file\n", nSourceId);
        DemoLogf("  byteCount=%d, cbWritten=%d\n", byteCount, cbWritten);
      } 
    }
  }
}


void DemoConnectBurstDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset 
    of a Highway) with a data rate between 8 kbit/s and 8.192Mbit/s. A pipe is 
    a characterized by its capacity (in 8 kbit/s increments). The user can create
    a pipe with the driver function <f OtxDrvCreatePipe>. A pipe is connected 
    to physical or logical devices and mapped to specific time-slots with the 
    driver function <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  // Godar board does not have any logical devices to support.
  // No pipe connection are needed at this time.
  return;
}


/*******************************************************************************

  @func
    Remove pipe connections

*******************************************************************************/
void DemoDisconnectBurstDevices(void)
{
#if 0
  OTX_RESULT otxRc;

  if(g_hBurstPipe != OTX_INVALID_HANDLE_VALUE) {
    otxRc = OtxDrvDisconnectPipe(g_hBurstPipe, OTX_PIPE_INPUT);
    if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
      DemoCheckRc(otxRc);
    }
    otxRc = OtxDrvDisconnectPipe(g_hBurstPipe, OTX_PIPE_OUTPUT);
    if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
      DemoCheckRc(otxRc);
    }

    // Close the pipe
    otxRc = OtxDrvClose(g_hBurstPipe); 
    DemoCheckRc(otxRc);
  }

  if(g_hTemuxPipe != OTX_INVALID_HANDLE_VALUE) {
    otxRc = OtxDrvDisconnectPipe(g_hTemuxPipe, OTX_PIPE_INPUT);
    if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
      DemoCheckRc(otxRc);
    }
    otxRc = OtxDrvDisconnectPipe(g_hTemuxPipe, OTX_PIPE_OUTPUT);
    if(OTX_RESULT_CODE(otxRc) != OTX_E_NOT_CONNECTED) {
      DemoCheckRc(otxRc);
    }

    // Close the pipe
    otxRc = OtxDrvClose(g_hTemuxPipe); 
    DemoCheckRc(otxRc);
  }
#else
  // Godar board does not have any logical devices to support.
  // No pipe connectio are needed at this time.
  return;
#endif
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoBurstIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_INT32 i;
  OTX_RESULT otxRc;

  while (g_bDemoRunnig) {
    OtxDrvWaitForSingleEvent(g_hNotificationBurstEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 nHighway;

      otxRc = OtxDrvGetEventDataEx(g_hBurstEventQueue, 
                                   sqEventData, 
                                   DEMO_EVENT_DATA_BUF_SIZE, 
                                   &coEventsFetched); 

      if(OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        // We have a driver event (one at least)

        if(coEventsFetched > DEMO_EVENT_DATA_BUF_SIZE) {
          // Buffer is not that big
          DemoLogf("ERROR: Event fetched %d events buffer holds %d\n", 
                   coEventsFetched,
                   DEMO_EVENT_DATA_BUF_SIZE);

          coEventsFetched = DEMO_EVENT_DATA_BUF_SIZE;
        }

        i = 0;
        g_FetchCount = coEventsFetched;
        g_FetchCountCur = i;
        while(i < coEventsFetched) {
          switch (sqEventData[i].m_eDeviceType) {
            case OTX_DEVICE_BURST:
              switch (sqEventData[i].m_nCode) {
              case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
                nHighway = sqEventData[i].m_nRequestId;
                // Check to make sure it is an event from one of the highways
                // that contain data being processed.
                if((nHighway < 0) || (nHighway >= g_nActiveHighways)) {
                  // if it is not an active highway just get the next event
                  if ((nHighway < 0) || (nHighway >= DEMO_T1_MAX_HWY_NO)) {
                    // It not a valid highway display it and get the next event
                    DemoLogf("DemoBurstIdleFunction: Invalid Highway(%d)...\n",
                              nHighway);
                  }
                }
                else {
                  otxRc = OtxDrvSetEventDataEx(
                                    g_hBurstReceiveEventQueue[nHighway],
                                    sqEventData[i].m_eDeviceType,   
                                    sqEventData[i].m_nSourceId,       
                                    sqEventData[i].m_hDevice,         
                                    sqEventData[i].m_nCode,           
                                    sqEventData[i].m_nRequestId,    
                                    sqEventData[i].m_nParam,          
                                    sqEventData[i].m_nTimeStamp       
                                    );
                  if (OTX_RESULT_CODE(otxRc) != OTX_S_OK) {
                    DemoLogf("DemoIdleFunction RECV(%d) error lost burst event 0x%X\n",
                             nHighway,
                             otxRc);
                  }
                }
                break;
              case OTX_BURST_SENDER_EC_BUF_DEPLETED:

                DemoBurstSender(sqEventData[i].m_nRequestId,  // Highway
                                sqEventData[i].m_nParam); // How much to transfer
                break;
              case OTX_BURST_EC_BUFFER_OVERRUN:
                 DemoLogf("Event Burst user buffer over/under run(%d)\n", 
                          sqEventData[i].m_nParam);
                 break;
              case OTX_BURST_EC_HWAY_BUFFER_OVERRUN:
                 DemoLogf("Event Burst user buffer over run highway %d\n", 
                          sqEventData[i].m_nParam);
                 break;
              case OTX_BURST_EC_HWAY_BUFFER_UNDERRUN:
                 DemoLogf("Event Burst user buffer under highway %d\n", 
                          sqEventData[i].m_nParam);
                 break;

              default:
                 DemoLogf("Event Burst user event (%d)\n", 
                          sqEventData[i].m_nCode);
                break;
              }
              break;
            default:
              // Unhandled event
              DemoLogf("Event from an unknown object 0x%X 0x%X\n", 
                       sqEventData[i].m_hDevice, 
                       sqEventData[i].m_eDeviceType);
              break;
          }
          i++;
          g_FetchCountCur = i;

        } // while()
      }
    } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
  }

  if(OTX_RESULT_CODE(otxRc) != OTX_S_OK) {
    DemoLogf("DemoIdleFunction:: Event OTX error 0x%X\n", otxRc);
  }
  else {
    DemoLogf("DemoIdleFunction:: thread done\n");
  }

  return 0;
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 
  // (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds 
  // is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = (nTimeStamp.nUint64 / 10000) - 
                         (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Prints a Driver Event. 
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TEMUX:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, 
                                   pEventData->m_nCode, 
                                   256, 
                                   &msgStr[0]);
    if(OTX_FAILURE(otxRc)) {
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);
     DemoLogf("\n"); 
     DemoLogf("Temux[%d] Event: code %ld paramater %ld\n", 
               pEventData->m_nSourceId, 
               pEventData->m_nCode,
               pEventData->m_nParam);
    }

    if(pEventData->m_nCode == OTX_TEMUX_EC_E1T1_SLP_R) {
      // Slips are detected in the E1/T1 when the system
      // is not receiving data. Which fills the display
      // with information that is not needed. Just keep
      // a count of the event. Display only the first one.
      if(OTX_SUCCESS(otxRc) && (g_nSlipCnt == 0)) {
        DemoLogf("Temux[%d] Event: %s (%d)\n", 
                  pEventData->m_nSourceId, 
                  msgStr, 
                  pEventData->m_nParam);
      }
      g_nSlipCnt++;
    }
    else {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);

      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("Temux[%d] Event: %s (0x%X)\n", 
                  pEventData->m_nSourceId, 
                  msgStr, 
                  pEventData->m_nParam);
      }
    }
    break;
  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, 
                                   pEventData->m_nCode, 
                                   256, 
                                   &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", 
                pEventData->m_nSourceId, 
                msgStr, 
                pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", 
               pEventData->m_nSourceId, 
               pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT otxRc;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      otxRc = OtxDrvGetEventDataEx(g_hEventQueue, 
                                   sqEventData, 
                                   DEMO_EVENT_DATA_BUF_SIZE, 
                                   &coEventsFetched); 

      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        // We have a driver event (one at least)

        if(coEventsFetched > DEMO_EVENT_DATA_BUF_SIZE) {
          // Buffer is not that big
          DemoLogf("ERROR: Event fetched %d events buffer holds %d\n", 
                   coEventsFetched,
                   DEMO_EVENT_DATA_BUF_SIZE);

          coEventsFetched = DEMO_EVENT_DATA_BUF_SIZE;
        }
        i = 0;
        while(i < coEventsFetched) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
          i++;
        } // while()
      }
    } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
  }

  return 0;
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT otxRc;

  if(commandChar > 0x20) {
    DemoLogf("\n Command > %c \n", commandChar);
  }
  else {
    DemoLogf("\n");
  }

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("\n");
    DemoLogf("\nCmd Synopsis, Active Highways %d\n", g_nActiveHighways);
    DemoLogf("===   ========\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");

    DemoLogf(" d    Display TEMUX configuration\n");
    DemoLogf(" D    Configure TEMUX device\n");

    DemoLogf(" g    Enable (start) burst GAMMA test\n");
    DemoLogf(" G    Clear Burst test status counters\n");
    DemoLogf(" e    Enable (start) burst fixed data test\n");
    DemoLogf(" E    Disable burst a single highway\n");
    DemoLogf(" f    Stop (abort) burst data transfer\n");
    DemoLogf(" R    Start burst receiver prompt for file name\n");
    DemoLogf(" s    Print GAMMA/Fixed test results for current\n");
    DemoLogf(" T    Set number of active highways\n");

    DemoLogf(" a    Send Debug command to device g_hBoard\n");
    DemoLogf(" A    Send Debug command to device g_hTemux\n");
    DemoLogf(" b    Send Debug command to device g_hBurst\n");
    DemoLogf(" J    Read 32 bit register from board device\n");
    DemoLogf(" K    Write 32 bit register to board device\n");
    DemoLogf(" j    Read 32 bit register from Temux device\n");
    DemoLogf(" k    Write 32 bit register to Temux device\n");
    DemoLogf(" l    Temux Line Status\n");
    DemoLogf(" t    Temux tributary Status\n");
    DemoLogf(" L    Read Temux Registers\n");
    DemoLogf(" v    Toggle verbose\n");
    DemoLogf(" V    Toggle Burst verbose\n");
    if(g_nSlipCnt) {
      DemoLogf(" C    Clear E1/T1 slip error count %d\n", g_nSlipCnt);
    }
    break;

  case 'C':
    {
      DemoLogf("E1/T1 slip error count %d is now cleared\n", g_nSlipCnt);
      g_nSlipCnt = 0;
    }
    break;

  case 'a':
    {
      char strBuf[12];
      static OTX_UINT32 nCmd = 0;

      DemoLogf("Debug command for device g_hBoard 0x%X\n", g_hBoard);

      DemoLogf("  Command  2 dump FPGA Control registers\n");
      DemoLogf("  Command  4 dump BURST DMA Buffer\n");
      DemoLogf("  Command 10 Fixed data\n");
      DemoLogf("  Command 11 GAMMA data\n");
      DemoLogf("  Command 12 HDLC data\n"); // used for this demo
      DemoLogf("  Command 13 TEMUX data\n"); // Data from TEMUX

      DemoLogf("Enter command (%d) 0=no command sent\n", nCmd);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }
      if(nCmd) {
        OtxDrvDbgDump(g_hBoard, nCmd);
      }
      DemoLogf("Done\n");
    }
    break;

  case 'A':
    {
      char strBuf[12];
      static OTX_UINT32 nCmd = 0;

      DemoLogf("Debug command for device g_hTemux 0x%X\n", g_hTemux);

      DemoLogf("Enter command (%d) 0=no command sent\n", nCmd);

      DemoLogf("  Command  1 Dump Temux registers\n");
      DemoLogf("  Command  2 Dump Temux interrupt registers\n");
      DemoLogf("  Command  3 Receive Clock Enable\n"); 
      DemoLogf("  Command  4 Line loopback Enable\n"); 
      DemoLogf("  Command  5 DEMAPPER loopback Enable\n"); 
      DemoLogf("  Command  6 MAPPER loopback Enable\n"); 
      DemoLogf("  Command  7 T1/E1 TRIBUTARY loopback Enable\n"); 
      DemoLogf("  Command  8 T1/E1 RXPCSC and TXPCSC loopback Enable\n"); 
      DemoLogf("  Command 10 Read interrupt status registers\n"); 
      DemoLogf("  Command 11 Display driver interrupt status\n"); 
      DemoLogf("  Command 12 DS3 status\n"); 
      DemoLogf("  Command 13 LO path status\n"); 

      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }
      if(nCmd) {
        OtxDrvDbgDump(g_hTemux, nCmd);
      }
      DemoLogf("Done\n");
    }
    break;

  case 'b':
    {
      char strBuf[12];
      static OTX_UINT32 nCmd = 0;

      DemoLogf("Debug command for device g_hBurst 0x%X\n", g_hBurst);

      DemoLogf("Enter command (%d) 0=no command sent\n", nCmd);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nCmd = strtoul(strBuf, NULL, 0);
      }
      if(nCmd) {
        OtxDrvDbgDump(g_hBurst, nCmd);
      }
      DemoLogf("Done\n");
    }
    break;

  case 'd':
    DemoTemuxGetDeviceInfo();
    break;

  case 'D':
    {
      char strBuf[12];
      OTX_BOOL    bTemuxConfig;
      static OTX_UINT32 nTemuxConfig = 0;

      DemoLogf(" 0  = exit command (Abort)\n");
      DemoLogf(" 1  = OTX_TEMUX_DEVICE_DEFAULT \n");
      DemoLogf(" 3  = OTX_TEMUX_T1VT15STS3CFG \n");
      DemoLogf(" 4  = OTX_TEMUX_T1VT2STS3CFG \n");
      DemoLogf(" 5  = OTX_TEMUX_E1VT2STS3CFG \n");
      DemoLogf(" 6  = OTX_TEMUX_DS3G747CFG \n");
      DemoLogf(" 7  = OTX_TEMUX_T1TU11AU4STM1CFG \n");
      DemoLogf(" 8  = OTX_TEMUX_T1TU12AU4STM1CFG \n");
      DemoLogf(" 9  = OTX_TEMUX_E1TU12AU4STM1CFG \n");
      DemoLogf("10  = OTX_TEMUX_T1TU11AU3STM1CFG \n");
      DemoLogf("11  = OTX_TEMUX_T1TU12AU3STM1CFG \n");
      DemoLogf("12  = OTX_TEMUX_E1TU12AU3STM1CFG \n");
      DemoLogf("13  = OTX_TEMUX_T1TU11AU3STM4CFG \n");
      DemoLogf("14  = OTX_TEMUX_E1TU12AU4STM4CFG \n");
      DemoLogf("15  = OTX_TEMUX_T1TU11AU4STM4CFG \n");
      DemoLogf("16  = OTX_TEMUX_DS3UNCHANNELIZEDCFG \n");
      DemoLogf("19  = OTX_TEMUX_DS3UNCHAU3STM1CFG \n");
      DemoLogf("20  = OTX_TEMUX_T1VT15STS3LPTIMINGCFG \n");
      DemoLogf("100 = Demo Defined 1 (E1)\n");
      DemoLogf("101 = Demo Defined 2 (T1)\n");

      DemoLogf("\nEnter configuration number (%d): ", nTemuxConfig);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTemuxConfig = strtoul(strBuf, NULL, 0);
      }

      if(nTemuxConfig > 99) {
        otxRc = OtxDrvDisable(g_hTemux, OTX_TASK_SYNC, 5000L); 
        DemoCheckDevRc(g_hTemux, otxRc);

        if(nTemuxConfig == 100) {
          DemoTemuxInterfaceUserDefinedE1();
        }
        if(nTemuxConfig == 101) {
          DemoTemuxInterfaceUserDefinedT1();
        }

        otxRc = OtxDrvEnable(g_hTemux, OTX_TASK_SYNC, 5000L); 
        DemoCheckDevRc(g_hTemux, otxRc);
      }
      else
      if(nTemuxConfig) {
        bTemuxConfig = DemoDisplayConfiguration(nTemuxConfig);
        if(bTemuxConfig == OTX_FALSE) {
          DemoLogf("Temux configuration not supported %d\n", nTemuxConfig);
        }

        otxRc = OtxDrvDisable(g_hTemux, OTX_TASK_SYNC, 5000L); 
        DemoCheckDevRc(g_hTemux, otxRc);

        DemoLogf("Temux configuration will take a few second\n");

        otxRc = OtxTemuxDeviceCfg(g_hTemux, 
                                  OTX_TEMUX_INTERFACE_HMVIP, 
                                  nTemuxConfig);
        if (OTX_FAILURE(otxRc)) {
          DemoLogf("Temux setup failed %d\n", nTemuxConfig);
          DemoLogf("ERROR(%#lX [%d.%d])\n", 
            OTX_RESULT_CODE(otxRc), 
            OTX_RESULT_SOURCE_ID(otxRc), 
            OTX_RESULT_LINE_NUM(otxRc));
        }

        otxRc = OtxDrvEnable(g_hTemux, OTX_TASK_SYNC, 5000L); 
        DemoCheckDevRc(g_hTemux, otxRc);
      }
      else {
        DemoLogf("Abort configuration update\n");
      }

      DemoLogf("Done\n");
    }
    break;

  case 'J':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nBdRdAddress = 0;

      DemoLogf("Read from 32 bit register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nBdRdAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nBdRdAddress = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvReadReg32(g_hBoard, nBdRdAddress, &readVal); 
      DemoCheckDevRc(g_hBoard, otxRc);

      DemoLogf("hBoard Address 0x%.4X = 0x%.8X\n", 
               nBdRdAddress, 
               (OTX_UINT32)readVal);
    }
    break;

  case 'K':
    {
      char strBuf[12];
      static OTX_UINT32 nBrWrAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to 32 bit register in the g_hBoard device\n");
      DemoLogf("Which Address (0x%X): ", nBrWrAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nBrWrAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvWriteReg32(g_hBoard, nBrWrAddress, nValue); 
      DemoCheckDevRc(g_hBoard, otxRc);

      DemoLogf("Wrote 0x%.8X to hBoard Address 0x%.4X\n", 
               nValue, 
               nBrWrAddress);
    }
    break;

  // Temux register access
  case 'j':
    {
      OTX_UINT32 nReadVal;
      OTX_UINT32 nIndex;
      char strBuf[12];
      static OTX_UINT32 nTmxRdAddress = 0;

      DemoLogf("Read from 32 bit register from the Temux device\n");
      DemoLogf("Which register offset (0x%X): ", nTmxRdAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTmxRdAddress = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvReadReg32(g_hTemux, nTmxRdAddress, &nReadVal); 
      DemoCheckDevRc(g_hTemux, otxRc);

      nIndex = nTmxRdAddress/4;
      DemoLogf("Temux Register index %d Address 0x%.4X = 0x%.8X\n", 
               nIndex, 
               nTmxRdAddress, 
               nReadVal);
    }
    break;

  case 'k':
    {
      char strBuf[12];
      OTX_UINT32 nIndex;
      static OTX_UINT32 nTmxWrAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to 32 bit register from the Temux device\n");
      DemoLogf("Which register offset  (0x%X): ", nTmxWrAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nTmxWrAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      otxRc = OtxDrvWriteReg32(g_hTemux, nTmxWrAddress, nValue); 
      DemoCheckDevRc(g_hTemux, otxRc);

      nIndex = nTmxWrAddress/4;
      DemoLogf("Register index %d Value 0x%.8X to Temux Address 0x%.4X\n", 
               nIndex, 
               nValue, 
               nTmxWrAddress);
    }
    break;

  case 'l':
    {
      OTX_UINT32 nLink;
      OtxTemuxGetLiStatusS  sLiStatus;

      if(g_nLink == 0) {
        // For now try to get all the links
        g_nLink = 1;
        DemoLogf("Need to get Temux configuration for number of links\n");
      }

      for(nLink=1; nLink <= g_nLink; nLink++) {
        otxRc = OtxTemuxGetLiStatus(g_hTemux, nLink, &sLiStatus);
        DemoCheckDevRc(g_hTemux, otxRc);
        if (OTX_SUCCESS(otxRc)) {
          if(sLiStatus.bLineStatus) {
            // Status is good
            DemoLogf("Link %d is OK\n", nLink);
          }
          else {
            DemoLogf("Link %d error detected:\n", nLink);
            if(sLiStatus.bLOS) {
              DemoLogf("  Loss of signal (LOS)\n");
            }
            if(sLiStatus.bLOF) {
              DemoLogf("  Loss of frame sync (LOF)\n");
            }
            if(sLiStatus.bTIM) {
              DemoLogf("  Trace identifier mismatch (TIM)\n");
            }
            if(sLiStatus.bSF) {
              DemoLogf("  Signal failure (SF)\n");
            }
            if(sLiStatus.bSD) {
              DemoLogf("  Signal degrade (SD)\n");
            }
            if(sLiStatus.bAIS) {
              DemoLogf("  Alarm indication signal defect (AIS)\n");
            }
          }
        }
        else {
          DemoLogf("Link %d status is not valid\n", nLink);
        }
      }
    }
    break;

  case 't':
    {
      OTX_UINT32 nLink;
      OTX_UINT32 nSpeSlice;
      OTX_UINT32 nTributary;
#ifdef DEMO_USE_T1 // T1
#else  // E1
      OtxTemuxGetE1S  sE1Status;
#endif

      if(g_nLink == 0) {
        // For now try to get all the links
        g_nLink = 1;
        DemoLogf("Need to get Temux configuration for number of links\n");
      }

#ifdef DEMO_USE_T1 // T1
#else  // E1

      for(nLink=1; nLink <= g_nLink; nLink++) {
        for(nSpeSlice=1; nSpeSlice <= OTX_TEMUX_MAX_STS1; nSpeSlice++) {
          for(nTributary=1; nTributary <= OTX_TEMUX_E1_TRIB; nTributary++) {
            otxRc = OtxTemuxGetE1Status(g_hTemux, 
                                        nLink, 
                                        nSpeSlice, 
                                        nTributary, 
                                        &sE1Status);
            DemoCheckDevRc(g_hTemux, otxRc);
            if(OTX_SUCCESS(otxRc)) {
              OtxTemuxFrameModeE eE1FramingMode;
              eE1FramingMode = g_sTemuxInformation.eFrameMode[nLink-1][nSpeSlice-1][nTributary-1];
              DemoLogf("Link%d SPE%d Tributary%d status:\n", 
                       nLink, nSpeSlice, nTributary);
              if(sE1Status.bTS16AISD) {
                DemoLogf("  incoming TS16 data stream has a low zero-bit density\n");
              }
              if(sE1Status.bAIS) {
                DemoLogf("  AIS all-ones condition has persisted for 100 ms\n");
              }
              if(sE1Status.bRED) {
                DemoLogf("  RED condition has persisted for 100 ms\n");
              }
              if(sE1Status.bAISD) {
                DemoLogf("  incoming data stream has a low zero-bit density\n");
              }
              if(sE1Status.bRMAI) {
                DemoLogf("  The Y bit has been logic one\n");
              }
              if(sE1Status.bRAI) {
                DemoLogf("  The A bit has been logic one\n");
              }
              // Useful if framing is CRC
              if((eE1FramingMode == OTX_TEMUX_E1_CRC) ||
                 (eE1FramingMode == OTX_TEMUX_E1_CAS_CRC)) {
                if(sE1Status.bCFEBE) {
                  DemoLogf("  CRC error E-bit is set high\n");
                }
                if(sE1Status.bRAICCRC) {
                  DemoLogf("  remote alarm A-bit high and the CRC error E-bit low\n");
                }
                if(sE1Status.bOOOF) {
                  DemoLogf("  CRC multi-frame framer searching for alignment\n");
                }
                if(sE1Status.bINCMF) {
                  DemoLogf("  CRC multi-frame alignment is framed\n");
                }
                else {
                  DemoLogf("  LOF CRC multi-frame alignment\n");
                }
              } 
              // Useful if framing is CAS
              if((eE1FramingMode == OTX_TEMUX_E1_CAS) ||
                 (eE1FramingMode == OTX_TEMUX_E1_CAS_CRC)) {
                if(sE1Status.bINSMF) {
                  DemoLogf("  signaling multi-frame alignment is framed\n");
                }
                else {
                  DemoLogf("  LOF signaling multi-frame alignment\n");
                }
              }
              if(sE1Status.bINF) {
                DemoLogf("  E1 is framed\n");
              }
              else {
                DemoLogf("  E1 frame alignment has been lost\n");
              }
            }
            else {
              DemoLogf("Link%d SPE%d Tributary%d status is not valid\n", 
                       nLink, nSpeSlice, nTributary);
            }
          }
        }
      }
#endif
    }
    break;

  case 'L':
    {
      OTX_UINT32 nReadVal;
      char strBuf[12];
      OTX_UINT32 nTmpIndex = 0;
      static OTX_UINT32 nTmxIndexStart = 0;
      static OTX_UINT32 nTmxIndexEnd = 0;

      DemoLogf("Display Temux registers starting with register index 0\n");
      DemoLogf("Start with register index (0x%X): ", nTmxIndexStart);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nTmxIndexStart = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("End with register index (0x%X): ", nTmxIndexEnd);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nTmxIndexEnd = strtoul(strBuf, NULL, 0);
      }

      nTmpIndex = nTmxIndexStart;
      while(nTmpIndex < nTmxIndexEnd) {
        otxRc = OtxDrvReadReg16(g_hTemux, nTmpIndex, &nReadVal); 
        if(OTX_FAILURE(otxRc)) {
          DemoLogf("Temux Read Failed 0x%X\n", otxRc);
          // Error, exit while loop 
          nTmpIndex = nTmxIndexEnd;
        }
        else {
          // TEMUX registers are 16 bits but they are accessed
          // on a 32 bit boundary and the MSB are not used.
          // Use register index as defined by PMC
          DemoLogf("Index 0x%.8X is 0x%.4X \n", nTmpIndex, nReadVal);
          nTmpIndex = nTmpIndex + 1;
        }
      }

    }
    break;

  case 'e':
    {
      OTX_UINT32 nReadVal;
      OTX_UINT32 nHwy;
      OTX_UINT32 nTs;
      OTX_UINT8  *pPattern;
      OTX_UINT32 nIndex = 0;
      // T1 OC-3 test equipment
      // 0x0 timeslots are not used in T1 data transfers from the TEMUX for 
      // data and TS0 is used for framing. (Unframed is not supported) 
      OTX_UINT8  nPatternT1Zero[32] = 
      {
        0x00,0x44,0,0x04,0x00,0x44,0,0x04,0x00,0x44,0,0x04,0x00,0x44,0,0x04,
        0x00,0x44,0,0x04,0x00,0x44,0,0x04,0x00,0x44,0,0x04,0x00,0x44,0,0x04
      };
      // Pattern generated by FPGA
      OTX_UINT8  nPatternFixed[32]  = 
      {
         0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
        16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31
      };
      OTX_INT32  nBurstHwy = 0;
      OTX_UINT32 nFpgaAddress = 0x0070;

      otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
      DemoCheckDevRc(g_hTemux, otxRc);

      g_nPattern = 0;

      // Get test pattern form hardware registers. The FPGA provides
      // fixed data patterns for the BURST device.
      otxRc = OtxDrvReadReg32(g_hBoard, nFpgaAddress, &nReadVal); 
      DemoCheckDevRc(g_hBoard, otxRc);

      if(nReadVal & 0x80000000) { // Fixed data pattern 0,1,2...31 each frame
        g_nPattern = 8;
      }
      if(nReadVal & 0x40000000) { // GAMMA data pattern 0-FF
        g_nPattern = 4;
        // Need to use GAMMA test option
        DemoLogf("GAMMA Pattern is enabled. Use GAMMA test not fixed\n");
      }
      if(nReadVal & 0x20000000) { // FISU data pattern 
        DemoLogf("FISU Pattern is enabled. Use HDLCEX to test not fixed\n");
        g_nPattern = 2;
      }

      if(g_nPattern == 8) {
        pPattern = &nPatternFixed[0];
      }
      else { // default to pattern 0
        if(g_sTemuxInformation.ePhdTribType[0][0] == OTX_TEMUX_E1) {
          // Note test equipment does not use one fixed pattern
          // for all highways so each highway has a different
          // data pattern.
          pPattern = &nPatternFixed[0];
          g_nPattern = 0xA;
        }
        else {
          pPattern = &nPatternT1Zero[0];
        }
      }

      // Load fixed data pattern
      for(nHwy = 0; nHwy < DEMO_MAX_HWY_NO; nHwy++) {
        for(nTs = 0; nTs < DEMO_BURST_BUF_SIZE; nTs++) {
          if(nIndex >= 32) {
            nIndex = 0;
          }
          g_sqSendFixedBuf[nHwy][nTs] = pPattern[nIndex];
          nIndex++;
        }
      }

      DemoLogf("Test pattern: \n");
      for(nTs = 0; nTs < 32; nTs++) {
        DemoLogf(" 0x%X", g_sqSendFixedBuf[0][nTs]);

        if(nTs == 16) {
          DemoLogf("\n"); 
        }
      }
      DemoLogf("\n");

      if(!g_hLogicDevs.mBurstReceiver[0].bGammaTestMode) {

        // If it is running stop the test
        if(g_hLogicDevs.mBurstReceiver[0].bFixedTestMode) {
          DemoBurstEndTest();
        }
        // Connect the burst device in loopback mode for testing.
        // The data sent should be received.
        DemoConnectBurstDevices();

        for(nBurstHwy = 0; nBurstHwy < g_nActiveHighways; nBurstHwy++) {
          // Requires all tributaries to be configured as E1 or T1.
          // E1 data uses 63 highways but the highways are not contunous.
          // That is HWY 0-20, HWY 28-48 and HWY 56-76 are used to store
          // the data. The demo does not enable the unused highways.
          if((g_sTemuxInformation.ePhdTribType[0][0] == OTX_TEMUX_E1) &&
             (((nBurstHwy >= 21) && (nBurstHwy < 28)) ||    // port 1 skip HWY
              ((nBurstHwy >= 49) && (nBurstHwy < 56)) ||    // port 1 skip HWY
              ((nBurstHwy >= 77) && (nBurstHwy < 84)) ||    // port 1 skip HWY
              ((nBurstHwy >= 106) && (nBurstHwy < 112)) ||  // port 2 skip HWY
              ((nBurstHwy >= 134) && (nBurstHwy < 140)) ||  // port 2 skip HWY
              ((nBurstHwy >= 162) && (nBurstHwy < 168))) ) {// port 2 skip HWY
            // E1 does not use this highway no need to configure it
            DemoLogf("E1 configuration highway %d not used\n", nBurstHwy);
          }
          else {
            DemoBurstTestFixedData(nBurstHwy);
          }
        }
      }
      else {
        DemoLogf("Burst GAMMA test is active\n");
      }
    }

    break;

  case 'g':
    {
      OTX_INT32 nBurstHwy;

      otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
      DemoCheckDevRc(g_hTemux, otxRc);

      if(!g_hLogicDevs.mBurstReceiver[0].bFixedTestMode) {

        // If it is running stop the test
        if(g_hLogicDevs.mBurstReceiver[0].bGammaTestMode) {
          DemoBurstEndTest();
        }

        // Connect the burst device in loopback mode for testing.
        DemoConnectBurstDevices();

        for(nBurstHwy = 0; nBurstHwy < g_nActiveHighways; nBurstHwy++) {
          // Requires all tributaries to be configured as E1 or T1.
          // E1 data uses 63 highways but the highways are not contunous.
          // That is HWY 0-20, HWY 28-48 and HWY 56-76 are used to store
          // the data. The demo does not enable the unused highways.
          if((g_sTemuxInformation.ePhdTribType[0][0] == OTX_TEMUX_E1) &&
             (((nBurstHwy >= 21) && (nBurstHwy < 28)) ||    // port 1 skip HWY
              ((nBurstHwy >= 49) && (nBurstHwy < 56)) ||    // port 1 skip HWY
              ((nBurstHwy >= 77) && (nBurstHwy < 84)) ||    // port 1 skip HWY
              ((nBurstHwy >= 106) && (nBurstHwy < 112)) ||  // port 2 skip HWY
              ((nBurstHwy >= 134) && (nBurstHwy < 140)) ||  // port 2 skip HWY
              ((nBurstHwy >= 162) && (nBurstHwy < 168))) ) {// port 2 skip HWY
            // E1 does not use this highway no need to configure it
            DemoLogf("E1 configuration highway %d not used\n", nBurstHwy);
          }
          else {
            DemoBurstTestGammaData(nBurstHwy);
          }
        }
      }
      else {
        DemoLogf("Burst fixed test is active\n");
      }
    }
    break;

  case 'G':  // Zero Gamma/Fixed Test Error Counters
    {
      OTX_UINT32 nBurstHwy;

      for(nBurstHwy = 0; nBurstHwy < DEMO_MAX_HWY_NO; nBurstHwy++) {
        g_hLogicDevs.mBurstReceiver[nBurstHwy].bTestReset = OTX_TRUE;
      }
        
      DemoLogf("Burst Receiver Test Results are now zero\n");
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      OtxDeviceStateE  eDevState;
      static OTX_UINT32 nRHighway = 0;

      DemoLogf("Enter highway number (%d): ", nRHighway);

      OsaConsoleGetStr(sqzBuf); 
      if(*sqzBuf) {
        nRHighway = strtoul(sqzBuf, NULL, 0);
      }

      if(nRHighway >= DEMO_MAX_HWY_NO){
        DemoLogf("Highway number %d out of range \n", nRHighway);
      }
      else
      if(g_hLogicDevs.mBurstReceiver[nRHighway].pRxFile == NULL) {
        DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", 
                 nRHighway);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
        }
        else {
          DemoLogf("Open file %s \n", sqzBuf);
          g_hLogicDevs.mBurstReceiver[nRHighway].pRxFile = fopen(sqzBuf, "wb");
        }

        if(g_hLogicDevs.mBurstReceiver[nRHighway].pRxFile == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
        }
        else {
          // Check for an already active transfer
          if(g_hLogicDevs.mBurstReceiver[nRHighway].bActive == OTX_FALSE) {
            // We always do all time slots
            g_hLogicDevs.mBurstReceiver[nRHighway].nTsToRecord = -1;
            g_hLogicDevs.mBurstReceiver[nRHighway].bActive = OTX_TRUE;

            otxRc = OtxDrvGetState(g_hBurst, &eDevState); 
            DemoCheckDevRc(g_hBurst, otxRc);

            if (eDevState == OTX_DEVICE_WORKING) {
              DemoLogf("Burst device is already enable\n");
            }
            else {
              // By enabling the Burst device, we start the process of 
              // transferring data from the hardware to the driver via 
              // DMA. The data is stored in the Burstdevice's buffer 
              // for processing later.
              DemoLogf("Enable g_hBurst\n");
              otxRc = OtxDrvEnable(g_hBurst, OTX_TASK_SYNC, 5000L);
              DemoCheckDevRc(g_hBurst, otxRc);
            }
          }
          else {
            DemoLogf("Receiver[%d] Thread already running.\n", nRHighway);
          }
        }
      }
      else {
        DemoLogf("Receiver file already in use\n");
      }
    }
    break;

  case 'f':
    {
      DemoBurstEndTest();
    }
    break;

  case 's':  // Test Results (condensed form)
    {
      OTX_INT32 nHwy;

      if(g_hLogicDevs.mBurstReceiver[0].bGammaTestMode) {
        DemoLogf("Burst Receiver Gamma Test Results: TS 0->32\n");
      }
      else {
        DemoLogf("Burst Receiver Fixed Data Test Results: TS 0->32\n");
      }

      for(nHwy = 0; nHwy < g_nActiveHighways; nHwy++) {
        if(g_hLogicDevs.mBurstReceiver[nHwy].bActive == OTX_TRUE) {
          if(g_hLogicDevs.mBurstReceiver[nHwy].bGammaTestMode ||
              g_hLogicDevs.mBurstReceiver[nHwy].bFixedTestMode) {
            if(g_hLogicDevs.mBurstReceiver[nHwy].nOverflowErrCnt != 0) {
              DemoLogf("Error: Burst[%d] overflow count %d\n", 
                       nHwy,
                       g_hLogicDevs.mBurstReceiver[nHwy].nOverflowErrCnt);

              g_hLogicDevs.mBurstReceiver[nHwy].nOverflowErrCnt = 0;
            }
            if(g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter == 0) {
              DemoLogf("Error: No burst data was received, Highway[%d]\n", 
                       nHwy);
            }
            else {
              OTX_UINT32 y;
              OTX_INT32 nErrCnt;
              // Used to verify event data is being received
              g_hLogicDevs.mBurstReceiver[nHwy].nDataBlockCounter = 0;

              for (y = 0; y < DEMO_BURST_TS_SIZE; y++) {
                nErrCnt = g_hLogicDevs.mBurstReceiver[nHwy].nGammaTestErrors[y];
                if ((y % 32) == 0) {
                  DemoLogf("HWY[%.2d] ", nHwy);
                }
                if ((y % 4) == 0) {
                  DemoLogf(" "); // Group by 4 timeslots to make it easier to read
                }
                if ((y % 8) == 0) {
                  DemoLogf(" "); // Group by 8 timeslots to make it easier to read
                }
                if (nErrCnt > 9) {
                  DemoLogf("*"); // Indicates a number larger than 9
                } else {
                  DemoLogf("%1d", nErrCnt);
                }
              }
              DemoLogf("\n");
            }
          } 
          else {
            DemoLogf("Test Mode is not enabled\n");
          }
        } 
        else {
          if(g_sTemuxInformation.ePhdTribType[0][0] == OTX_TEMUX_E1) {
            DemoLogf("Burst Receiver Thread Highway[%d] is not used with E1\n", 
                     nHwy);
          }
          else {
            DemoLogf("Burst Receiver Thread Highway[%d] is not active\n", 
                     nHwy);
          }
        }
      }
    }
    
    if(g_nSlipCnt) {
      DemoLogf("E1/T1 slip error count is %d\n", g_nSlipCnt);
    }

    break;

  case 'T':
    {
      char strBuf[12];
      static OTX_UINT32 nABH = DEMO_ACTIVE_HWY_NO;

      DemoLogf("Enter highway number (%d): ", nABH);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nABH = strtoul(strBuf, NULL, 0);
      }
      if(DEMO_MAX_HWY_NO > nABH) {
        g_nActiveHighways = nABH;

        DemoLogf("Number of active burst highway is %d\n", nABH);
      }
      else {
        DemoLogf("Highway number %d out of range\n", nABH);
      }
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'v':
    {
      char strBuf[12];

      g_bVerbose = !g_bVerbose;

      DemoLogf("Verbose set to %s\n", 
               g_bVerbose ? "Enabled" : "Disabled");

      DemoLogf("Enter highway number (%d): ", g_nTstHighway);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        g_nTstHighway = strtoul(strBuf, NULL, 0);
      }
      if(g_nTstHighway < DEMO_MAX_HWY_NO) {
        // Display highway error on this highway
        DemoLogf("Enable data error display highway (%d)\n", g_nTstHighway);

        g_hLogicDevs.mBurstReceiver[g_nTstHighway].bDisplayDataErr = OTX_TRUE;
      }
    } 
    break;

  case 'V':
    {
      char strBuf[12];

      g_bBurstVerbose = !g_bBurstVerbose;

      DemoLogf("Burst verbose set to %s\n", 
               g_bBurstVerbose ? "Enabled" : "Disabled");

      DemoLogf("Enter highway number (%d): ", g_nTstHighway);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        g_nTstHighway = strtoul(strBuf, NULL, 0);
      }
      if(g_nTstHighway < DEMO_MAX_HWY_NO) {
        // Display highway
        g_hLogicDevs.mBurstReceiver[g_nTstHighway].bVerbose = g_bBurstVerbose;
        g_hLogicDevs.mBurstSender[g_nTstHighway].bVerbose   = g_bBurstVerbose;
      }
    }
    break;

  case 'E':

    {
      char strBuf[12];
      OTX_BOOL bCurrent;
      static OTX_UINT32 nBH = 0;

      DemoLogf("Enter highway number (%d): ", nBH);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nBH = strtoul(strBuf, NULL, 0);
      }
      if(g_nActiveHighways > (OTX_INT32) nBH) {
        bCurrent = g_hLogicDevs.mBurstReceiver[nBH].bActive;

        g_hLogicDevs.mBurstReceiver[nBH].bActive =!bCurrent;

        DemoLogf("Burst highway%d is %s\n", 
                 nBH,       
                 !bCurrent ? "Enabled" : "Disabled");
      }
      else {
        DemoLogf("Highway number %d out of range\n", nBH);
      }
    }
    break;

  case 'q':
  case 'Q':
    DemoBurstEndTest();
    return 1;
    break;

  case 0xD: // CR
    // No command just return
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  char        commandChar;
  OTX_INT32   exitFlag = 0;
  OTX_RESULT  otxRc;
  OTX_UINT32  i;
  OTX_UINT32  x;
  OTX_BOOL    bTemuxConfig;
  OtxPhysicalDeviceTypeE nBoardType;

  for (i = 0; i < DEMO_MAX_HWY_NO; i++) {
    // Set each Burst test buffer to a unique value
    for (x = 0; x < (OTX_UINT32) DEMO_BURST_TS_SIZE; x++) {
      g_hLogicDevs.mBurstSender[i].sqGammaData[x] = (unsigned char) i + 1;
    }

    // Clear error counters
    g_hLogicDevs.mBurstSender[i].nGammaZeroXferCnt = 0;

    for (x = 0; x < (OTX_UINT32) DEMO_BURST_TS_SIZE; x++) {
      g_hLogicDevs.mBurstReceiver[i].nGammaTestErrors[x] = 0;
    }

    g_hLogicDevs.mBurstReceiver[i].nOverflowErrCnt = 0;

    g_hBurstReceiveEvent[i] = OTX_INVALID_EVENT_VALUE;
    g_hBurstReceiveEventQueue[i] = 0;
  }

  printf("Demo App for Godar PCIe, %s, Odin TeleSystems Inc.\n", __DATE__);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen("GodarPCIeDemo.log", "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", "GodarPCIeDemo.log");
    OsaConsoleExit();
    exit(-1);
  }

  otxRc = OtxDrvConnectLib(); 
  DemoCheckRc(otxRc);

  bTemuxConfig = DemoDisplayConfiguration((OTX_UINT32) g_demoEnv.eTemuxConfig);
  if(bTemuxConfig == OTX_FALSE) {
    DemoLogf("Temux configuration not supported %d\n", g_demoEnv.eTemuxConfig);
    OsaConsoleExit();
    exit(-1);
  }

  // Create the event queue to be used by the driver to inform the application
  // about events
  otxRc = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(otxRc) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;  
    otxRc = OTX_S_OK;
  }
  DemoCheckRc(otxRc);

  // Create Event Queue for driver notifications
  otxRc = OtxDrvCreateEventQueue(128, 
                                 g_hNotificationEvent, 
                                 &g_hEventQueue); 
  DemoCheckRc(otxRc);

  // Enable Event Queue
  otxRc = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 0); 
  DemoCheckRc(otxRc);


  // Create an event queue just for burst 
  otxRc = OtxDrvCreateEvent(&g_hNotificationBurstEvent);
  if (OTX_RESULT_CODE(otxRc) == OTX_E_NOTIMPL) {
    g_hNotificationBurstEvent = (OTX_EVENT) NULL;   
    otxRc = OTX_S_OK;
  }
  DemoCheckRc(otxRc);

  // Create Event Queue for driver notifications
  otxRc = OtxDrvCreateEventQueue(2048, 
                                 g_hNotificationBurstEvent, 
                                 &g_hBurstEventQueue); 
  DemoCheckRc(otxRc);

  // Enable Event Queue
  otxRc = OtxDrvEnable(g_hBurstEventQueue, OTX_TASK_SYNC, 0); 
  DemoCheckRc(otxRc);

  for (i = 0; i < DEMO_MAX_HWY_NO; i++) {
    // Create a seperate event queue for  each burst receive highway 
    // This will keep one highway from block the other while processing
    // the data from the highway. 
    otxRc = OtxDrvCreateEvent(&g_hBurstReceiveEvent[i]);
    if (OTX_RESULT_CODE(otxRc) == OTX_E_NOTIMPL) {
      g_hBurstReceiveEvent[i] = NULL;
      otxRc = OTX_S_OK;
    }
    else {
      DemoCheckRc(otxRc);
    }
    otxRc = OtxDrvCreateEventQueue(128, 
                                   g_hBurstReceiveEvent[i], 
                                   &g_hBurstReceiveEventQueue[i]); 
    DemoCheckRc(otxRc);

    // Enable Event Queue
    otxRc = OtxDrvEnable(g_hBurstReceiveEventQueue[i], OTX_TASK_SYNC, 0); 
    DemoCheckRc(otxRc);
  }


  // Open board
  nBoardType = (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType;
  otxRc = OtxDrvOpenPhysicalDevice(0, 
                                   nBoardType, 
                                   0, //g_demoEnv.boardNo
                                   0, 
                                   g_hEventQueue, 
                                   &g_hBoard); 
  DemoCheckRc(otxRc);

  otxRc = OtxDrvEnable(g_hBoard, OTX_TASK_SYNC, 5000L); 
  DemoCheckDevRc(g_hBoard, otxRc);


  DemoLogf("Open Temux Device\n");
  otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                   OTX_DEVICE_TEMUX, 
                                   0, 
                                   0, 
                                   g_hEventQueue, 
                                   &g_hTemux); 
  DemoCheckDevRc(g_hBoard, otxRc);


  DemoLogf("Open Burst Device\n");
  //
  // Setup Burst device
  //
  // Get a handle to the physical Burst device.
  otxRc = OtxDrvOpenPhysicalDevice(g_hBoard, 
                                   OTX_DEVICE_BURST, 
                                   0, 
                                   0, 
                                   g_hBurstEventQueue, 
                                   &g_hBurst); 
  DemoCheckDevRc(g_hBoard, otxRc);

  DemoLogf("Burst RX FIFO trigger is 0x%x (%d)\n", 
            DEMO_BURST_BUF_SIZE, 
            DEMO_BURST_BUF_SIZE);

  // set device buffer event trigger levels
  otxRc = OtxDrvSetAttributeValueUint32(g_hBurst, 
                                        OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  
                                        DEMO_BURST_BUF_SIZE); 
  DemoCheckDevRc(g_hBurst, otxRc);

  DemoLogf("Burst TX FIFO trigger is 0x%x (%d)\n", 
           DEMO_BURST_BUF_SIZE, 
           DEMO_BURST_BUF_SIZE);

  otxRc = OtxDrvSetAttributeValueUint32(g_hBurst, 
                                        OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    
                                        DEMO_BURST_BUF_SIZE); 
  DemoCheckDevRc(g_hBurst, otxRc);

  // Optionally (mainly for testing, or for improving system performance for single 
  // direction burst applications, disable one direction of the burst device
  if(g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupt is disabled!\n");
    otxRc = OtxDrvSetAttributeValueBool(g_hBurst, 
                                        OTX_ATTR_BURST_RECV_DISABLED,  
                                        OTX_TRUE); 
    DemoCheckDevRc(g_hBurst, otxRc);
  }
  if(g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupt is disabled!\n");
    otxRc = OtxDrvSetAttributeValueBool(g_hBurst, 
                                        OTX_ATTR_BURST_SEND_DISABLED,  
                                        OTX_TRUE); 
    DemoCheckDevRc(g_hBurst, otxRc);
  }

  // Configure pipe connection within driver.
  // Godar does not have TSS device so hardware is already connected
  // but the driver can require a pipe conection for some devices
  DemoConnectBurstDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with OtxDrvGetAttributeValue)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckDevRc(g_hBoard, otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckDevRc(g_hBoard, otxRc);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckDevRc(g_hBoard, otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckDevRc(g_hBoard, otxRc);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    otxRc = OtxDrvGetAttributeType(g_hBoard, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckDevRc(g_hBoard, otxRc);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    otxRc = OtxDrvGetAttributeValueSz(g_hBoard, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckDevRc(g_hBoard, otxRc);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }


  otxRc = OtxTemuxGetInfo(g_hTemux, &g_sTemuxInformation);
  DemoCheckDevRc(g_hTemux, otxRc);

  DemoLogf("Interface %s\n", g_sTemuxInformation.eHwInterface ? "HMVIP" : "SBI");

  // For now just display this information
  DemoLogf("Number of Lines supported %d\n", g_sTemuxInformation.nNumSpans);
  if((g_sTemuxInformation.nNumSpans < OTX_TEMUX_MAX_LINK) && 
     (g_sTemuxInformation.nNumSpans != 0)) {
    g_nLink = g_sTemuxInformation.nNumSpans;
  }
  else {
    DemoLogf("Number of Lines supported is out of range\n");
    g_nLink = OTX_TEMUX_MAX_LINK;
  }

  DemoLogf("Configure Temux device, this will take a few seconds\n");
  if(g_demoEnv.eTemuxConfig) {
    // Use one of the Temux configuration from the API
    otxRc = OtxTemuxDeviceCfg(g_hTemux, 
                              OTX_TEMUX_INTERFACE_HMVIP, 
                              g_demoEnv.eTemuxConfig);
    if (OTX_FAILURE(otxRc)) {
      DemoLogf("Temux setup failed %d\n", g_demoEnv.eTemuxConfig);
      DemoLogf("ERROR(%#lX [%d.%d])\n", 
        OTX_RESULT_CODE(otxRc), 
        OTX_RESULT_SOURCE_ID(otxRc), 
        OTX_RESULT_LINE_NUM(otxRc));
    }
  }
  else {
#if DEMO_USE_T1
    // Use the T1 configuration from this file
    DemoTemuxInterfaceUserDefinedT1();
#else
    // Use the E1 configuration from this file
    DemoTemuxInterfaceUserDefinedE1();
#endif
  }

  otxRc = OtxDrvEnable(g_hTemux, OTX_TASK_SYNC, 5000L); 
  DemoCheckDevRc(g_hTemux, otxRc);

  OsaThreadStart(DemoIdleFunction);

  for (i = 0; i < DEMO_MAX_HWY_NO; i++) {
    g_nHighwayNumber[i] = i; // pass a static global value to thread
    OsaThreadStartEx(DemoBurstReceiveProc, (void *) &g_nHighwayNumber[i]);
  }
  // Let the threads that will be processing the data start
  OsaThreadSleep(100);

  OsaThreadStart(DemoBurstIdleFunction);        
  // Let the burst event threads start
  OsaThreadSleep(1000);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Close device that may have been created
  DemoDisconnectBurstDevices();

  OsaThreadSleep(100);

  g_bDemoRunnig = OTX_FALSE;

  // Give the threads a chance to exit
  OsaThreadSleep(1000);

  OtxDrvReleaseLib();

  return 0;
}


/*******************************************************************************

  @func

  Displays the device configuration being used for this device.
  The appication can select one of the predefined configuration or
  use a user defined configuration.

  @comm
  returns OTX_FALSE if configuration is unknown
*******************************************************************************/
static OTX_BOOL DemoDisplayConfiguration(OTX_UINT32 nConfig)
{
  OTX_BOOL bConfigOk = OTX_TRUE;

  switch(nConfig) {
  case 0: // User defined for custom configuration
    DemoLogf("Device Configuration File-0  = User Defined\n");
    break;

  case OTX_TEMUX_DEVICE_DEFAULT: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-1  = OTX_TEMUX_DEVICE_DEFAULT \n");
    break;

  case OTX_TEMUX_DS3M13CFG: // See OtxTemuxDevConfigureE
    // Not supported by the board
    DemoLogf("Device Configuration File-2  = OTX_TEMUX_DS3M13CFG \n");
    break;

  case OTX_TEMUX_T1VT15STS3CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-3  = OTX_TEMUX_T1VT15STS3CFG \n");
    break;

  case OTX_TEMUX_T1VT2STS3CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-4  = OTX_TEMUX_T1VT2STS3CFG \n");
    break;

  case OTX_TEMUX_E1VT2STS3CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-5  = OTX_TEMUX_E1VT2STS3CFG \n");
    break;

  case OTX_TEMUX_DS3G747CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-6  = OTX_TEMUX_DS3G747CFG \n");
    break;

  case OTX_TEMUX_T1TU11AU4STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-7  = OTX_TEMUX_T1TU11AU4STM1CFG \n");
    break;

  case OTX_TEMUX_T1TU12AU4STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-8  = OTX_TEMUX_T1TU12AU4STM1CFG \n");
    break;

  case OTX_TEMUX_E1TU12AU4STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-9  = OTX_TEMUX_E1TU12AU4STM1CFG \n");
    break;

  case OTX_TEMUX_T1TU11AU3STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-10  = OTX_TEMUX_T1TU11AU3STM1CFG \n");
    break;

  case OTX_TEMUX_T1TU12AU3STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-11  = OTX_TEMUX_T1TU12AU3STM1CFG \n");
    break;

  case OTX_TEMUX_E1TU12AU3STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-12  = OTX_TEMUX_E1TU12AU3STM1CFG \n");
    break;

  case OTX_TEMUX_T1TU11AU3STM4CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-13  = OTX_TEMUX_T1TU11AU3STM4CFG \n");
    break;

  case OTX_TEMUX_E1TU12AU4STM4CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-14  = OTX_TEMUX_E1TU12AU4STM4CFG \n");
    break;

  case OTX_TEMUX_T1TU11AU4STM4CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-15  = OTX_TEMUX_T1TU11AU4STM4CFG \n");
    break;

  case OTX_TEMUX_DS3UNCHANNELIZEDCFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-16  = OTX_TEMUX_DS3UNCHANNELIZEDCFG \n");
    break;

  case OTX_TEMUX_DS3M13AU4STM1CFG: // See OtxTemuxDevConfigureE
    // Not supported by the board
    DemoLogf("Device Configuration File-17  = OTX_TEMUX_DS3M13AU4STM1CFG \n");
    break;

  case OTX_TEMUX_DS3M13AU3STM1CFG: // See OtxTemuxDevConfigureE
    // Not supported by the board
    DemoLogf("Device Configuration File-18  = OTX_TEMUX_DS3M13AU3STM1CFG \n");
    break;

  case OTX_TEMUX_DS3UNCHAU3STM1CFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-19  = OTX_TEMUX_DS3UNCHAU3STM1CFG \n");
    break;

  case OTX_TEMUX_T1VT15STS3LPTIMINGCFG: // See OtxTemuxDevConfigureE
    DemoLogf("Device Configuration File-20  = OTX_TEMUX_T1VT15STS3LPTIMINGCFG \n");
    break;

  default:
    // Unknown configuration for the Temux device
    DemoLogf("Error:: Temux configuration not defined %d\n", nConfig);
    bConfigOk = OTX_FALSE;
    break;
  }

  return(bConfigOk);
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /T[x] : Board Type used for the demo, where\n");
  DemoLogf("                    1  = Use Godar-1-PCIe Board\n");
  DemoLogf("                    2  = Use Godar-2-PCIe Board\n");
  DemoLogf(" /D[x] : Temux config OtxTemuxDevConfigureE\n");
  DemoLogf("                    0  = User Defined\n");
  DemoLogf("                    1  = OTX_TEMUX_DEVICE_DEFAULT \n");
  DemoLogf("                    3  = OTX_TEMUX_T1VT15STS3CFG \n");
  DemoLogf("                    4  = OTX_TEMUX_T1VT2STS3CFG \n");
  DemoLogf("                    5  = OTX_TEMUX_E1VT2STS3CFG \n");
  DemoLogf("                    6  = OTX_TEMUX_DS3G747CFG \n");
  DemoLogf("                    7  = OTX_TEMUX_T1TU11AU4STM1CFG \n");
  DemoLogf("                    8  = OTX_TEMUX_T1TU12AU4STM1CFG \n");
  DemoLogf("                    9  = OTX_TEMUX_E1TU12AU4STM1CFG \n");
  DemoLogf("                   10  = OTX_TEMUX_T1TU11AU3STM1CFG \n");
  DemoLogf("                   11  = OTX_TEMUX_T1TU12AU3STM1CFG \n");
  DemoLogf("                   12  = OTX_TEMUX_E1TU12AU3STM1CFG \n");
  DemoLogf(" Note: the following set User Defined OtxTemuxDevConfigureE = 0\n");
  DemoLogf(" /E[x] : Temux timing mode OtxTemuxTxTiming\n");
  DemoLogf("                    1  = OTX_TEMUX_TX_TT_LOOP (Default)\n");
  DemoLogf("                    2  = OTX_TEMUX_TX_TT_CTCLK \n");
  DemoLogf("                    3  = OTX_TEMUX_TX_TT_DSSI \n");
  DemoLogf("                    4  = OTX_TEMUX_TX_TT_TISS \n");
  DemoLogf("                    5  = OTX_TEMUX_TX_TT_MUX \n");
  DemoLogf(" /F[x] : Temux T1/E1 tributary mapping OtxTemuxMapTribMode\n");
  DemoLogf("                    1  = OTX_TEMUX_BIT_ASYN (Default)\n");
  DemoLogf("                    2  = OTX_TEMUX_BYTE_SYNC \n");
  DemoLogf("                    3  = OTX_TEMUX_TVT \n");
  DemoLogf(" /G[x] : Temux framing mode OtxTemuxFrameModeE (E1)\n");
  DemoLogf("                    1  = OTX_TEMUX_E1_STANDARD (Default)\n");
  DemoLogf("                    2  = OTX_TEMUX_E1_CRC \n");
  DemoLogf("                    3  = OTX_TEMUX_E1_CAS \n");
  DemoLogf("                    4  = OTX_TEMUX_E1_CAS_CRC \n");
  DemoLogf(" /H[x] : Temux link clock source OtxTemuxClockSrcE\n");
  DemoLogf("                    1  = OTX_TEMUX_CLOCK_INTERNAL (Default)\n");
  DemoLogf("                    2  = OTX_TEMUX_CLOCK_LOOP \n");
  DemoLogf(" /Q    : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set defaults before reading from command line
  envP->nBoardType         = OTX_DEVICE_GODAR_2_PCI_EXPRESS;
  envP->pszInputStr        = NULL;
  envP->bExitOnError       = OTX_FALSE;
  envP->boardNo            = DEMO_DEFAULT_BOARD_NO;
  envP->coLi               = 2;

  envP->bDisableBurstTx    = OTX_TRUE; // This version is receive only 
  envP->bDisableBurstRx    = OTX_FALSE; // We want this enable 
  envP->eTemuxConfig       = OTX_TEMUX_E1VT2STS3CFG;
  envP->eTxTimingMode      = OTX_TEMUX_TX_TT_LOOP;
  envP->eLinkSrcOrDest     = OTX_TEMUX_BIT_ASYN;

  envP->eE1FramingMode     = OTX_TEMUX_E1_STANDARD;
  envP->eClkSrcMode        = OTX_TEMUX_CLOCK_INTERNAL;

  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);

  for (i = 1; i < argc; i++) {   // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_GODAR_2_PCI_EXPRESS;
            envP->coLi = 2;
          } else if (*(argv[i] + 2) == '1') {
            envP->nBoardType = OTX_DEVICE_GODAR_1_PCI_EXPRESS;
            envP->coLi = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'D') {   // Select Temux configuration
          envP->eTemuxConfig = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i] + 1) == 'E') {
          envP->eTemuxConfig = (OtxTemuxDevConfigureE) 0; // User defined setup
          // Temux timing mode
          if (*(argv[i] + 2) == '1') {
            envP->eTxTimingMode = OTX_TEMUX_TX_TT_LOOP;
          } else 
          if (*(argv[i] + 2) == '2') {
            envP->eTxTimingMode = OTX_TEMUX_TX_TT_CTCLK;
          } else 
          if (*(argv[i] + 2) == '3') {
            envP->eTxTimingMode = OTX_TEMUX_TX_TT_DSSI;
          } else 
          if (*(argv[i] + 2) == '4') {
            envP->eTxTimingMode = OTX_TEMUX_TX_TT_TISS;
          } else 
          if (*(argv[i] + 2) == '5') {
            envP->eTxTimingMode = OTX_TEMUX_TX_TT_MUX;
          } else {
            DemoLogf("Error: Unknown Temux timing mode switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'F') {
          envP->eTemuxConfig = (OtxTemuxDevConfigureE) 0; // User defined setup
          // Temux Tributary Map Mode
          if (*(argv[i] + 2) == '1') {
            envP->eLinkSrcOrDest = OTX_TEMUX_BIT_ASYN;
          } else 
          if (*(argv[i] + 2) == '2') {
            envP->eLinkSrcOrDest = OTX_TEMUX_BYTE_SYNC;
          } else 
          if (*(argv[i] + 2) == '3') {
            envP->eLinkSrcOrDest = OTX_TEMUX_TVT;
          } else {
            DemoLogf("Error: Unknown Temux Tributary Mode switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'G') {
          envP->eTemuxConfig = (OtxTemuxDevConfigureE) 0; // User defined setup
          // Temux T1/E1 framing mode
          if (*(argv[i] + 2) == '1') {
            envP->eE1FramingMode = OTX_TEMUX_E1_STANDARD;;
          } else 
          if (*(argv[i] + 2) == '2') {
            envP->eE1FramingMode = OTX_TEMUX_E1_CRC;;
          } else 
          if (*(argv[i] + 2) == '3') {
            envP->eE1FramingMode = OTX_TEMUX_E1_CAS;;
          } else 
          if (*(argv[i] + 2) == '4') {
            envP->eE1FramingMode = OTX_TEMUX_E1_CAS_CRC;;
          } else {
            DemoLogf("Error: Unknown Temux T1/E1 framing mode switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'H') {
          envP->eTemuxConfig = (OtxTemuxDevConfigureE) 0; // User defined setup
          // Temux clock Mode
          if (*(argv[i] + 2) == '1') {
            envP->eClkSrcMode = OTX_TEMUX_CLOCK_INTERNAL;;
          } else 
          if (*(argv[i] + 2) == '2') {
            envP->eClkSrcMode = OTX_TEMUX_CLOCK_LOOP;;
          } else {
            DemoLogf("Error: Unknown Temux clock mode switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}

